"""
Module for loading datasets
"""

import os
import sys
import threading
from collections import defaultdict, OrderedDict
from functools import reduce
from os.path import join

import h5py
import numpy as np
import scipy as sp
from PIL import Image
from tensorflow.contrib.learn.python.learn.datasets.mnist import read_data_sets

from boml import load_data as dl
from boml.load_data.datasets import MetaDataset
from boml.load_data.datasets.dl_utils import (
    get_rand_state,
    vstack,
    get_data,
    stack_or_concat,
    get_targets,
    as_list,
    get_indices_balanced_classes,
    test_if_balanced,
    to_one_hot_enc,
    merge_dicts,
    as_tuple_or_list,
)

try:
    from sklearn.datasets import make_classification, make_regression
except ImportError as _err:
    print(_err, file=sys.stderr)
    make_regression, make_classification = None, None

from_env = os.getenv("DATASETS_FOLDER")
if from_env:
    DATA_FOLDER = from_env
else:
    print(
        "Environment variable DATASETS_FOLDER not found. Variables HELP_WIN and HELP_UBUNTU contain info."
    )
    DATA_FOLDER = os.getcwd()
    _COMMON_BEGIN = (
        "You can set environment variable DATASETS_FOLDER to"
        "specify root folder in which you store various datasets. \n"
    )
    _COMMON_END = """\n
    You can also skip this step... \n
    In this case all load_* methods take a FOLDER path as first argument. \n
    Bye."""
    HELP_UBUNTU = (
        _COMMON_BEGIN
        + """
    Bash command is: export DATASETS_FOLDER='absolute/path/to/dataset/folder \n
    Remember! To add the global variable kinda permanently in your system you should add export command in
    bash.bashrc file located in etc folder, if you want to set it globally, or .bashrc in your home directory
    if you want to set it only locally.
    """
        + _COMMON_END
    )

    HELP_WIN = (
        _COMMON_BEGIN
        + """
    Cmd command is: Set DATASETS_FOLDER absolute/path/to/dataset/folder  for one session. \n
    To set it permanently use SetX instead of Set (and probably reboot system)
    """
        + _COMMON_END
    )

print("Data folder is", DATA_FOLDER)

# easy to find!
IRIS_TRAINING = os.path.join(DATA_FOLDER, "iris", "training.csv")
IRIS_TEST = os.path.join(DATA_FOLDER, "iris", "test.csv")
MNIST_DIR = os.path.join(DATA_FOLDER, "mnist_data")
CALTECH101_30_DIR = os.path.join(DATA_FOLDER, "caltech101-30")
CALTECH101_DIR = os.path.join(DATA_FOLDER, "caltech")
CENSUS_TRAIN = os.path.join(DATA_FOLDER, "census", "train.csv")
CENSUS_TEST = os.path.join(DATA_FOLDER, "census", "test.csv")
CIFAR10_DIR = os.path.join(DATA_FOLDER, "CIFAR-10")
CIFAR100_DIR = os.path.join(DATA_FOLDER, "CIFAR-100")
REALSIM = os.path.join(DATA_FOLDER, "realsim")

# scikit learn datasets
SCIKIT_LEARN_DATA = os.path.join(DATA_FOLDER, "scikit_learn_data")

IMAGENET_BASE_FOLDER = join(DATA_FOLDER, "imagenet")
MINI_IMAGENET_FOLDER = join(DATA_FOLDER, join("imagenet", "mini_v1"))
MINI_IMAGENET_FOLDER_RES84 = join(DATA_FOLDER, join("imagenet", "mini_res84"))
MINI_IMAGENET_FOLDER_V2 = join(DATA_FOLDER, join("imagenet", "mini_v2"))
MINI_IMAGENET_FOLDER_V3 = join(DATA_FOLDER, join("imagenet", "mini_v3"))

OMNIGLOT_RESIZED = join(DATA_FOLDER, "omniglot_resized")


def balanced_choice_wr(a, num, rand=None):
    rand = get_rand_state(rand)
    lst = [len(a)] * (num // len(a)) + [num % len(a)]
    return np.concatenate([rand.choice(a, size=(d,), replace=False) for d in lst])


def mnist(
    folder=None, one_hot=True, partitions=None, filters=None, maps=None, shuffle=False
):
    if not folder:
        folder = MNIST_DIR
    datasets = read_data_sets(folder, one_hot=one_hot)
    train = dl.Dataset(datasets.train.images, datasets.train.labels, name="MNIST")
    validation = dl.Dataset(
        datasets.validation.images, datasets.validation.labels, name="MNIST"
    )
    test = dl.Dataset(datasets.test.images, datasets.test.labels, name="MNIST")
    res = [train, validation, test]
    if partitions:
        res = redivide_data(
            res,
            partition_proportions=partitions,
            filters=filters,
            maps=maps,
            shuffle=shuffle,
        )
        res += [None] * (3 - len(res))
    return dl.Datasets.from_list(res)


def omni_light(folder=join(DATA_FOLDER, "omniglot-light"), add_bias=False):
    """
    Extract from omniglot dataset with rotated images, 100 classes,
    3 examples per class in training set
    3 examples per class in validation set
    15 examples per class in test set
    """
    file = h5py.File(os.path.join(folder, "omni-light.h5"), "r")
    return dl.Datasets.from_list(
        [
            dl.Dataset(
                np.array(file["X_ft_tr"]),
                np.array(file["Y_tr"]),
                info={"original images": np.array(file["X_orig_tr"])},
                add_bias=add_bias,
            ),
            dl.Dataset(
                np.array(file["X_ft_val"]),
                np.array(file["Y_val"]),
                info={"original images": np.array(file["X_orig_val"])},
                add_bias=add_bias,
            ),
            dl.Dataset(
                np.array(file["X_ft_test"]),
                np.array(file["Y_test"]),
                info={"original images": np.array(file["X_orig_test"])},
                add_bias=add_bias,
            ),
        ]
    )


load_omni_light = omni_light


class OmniglotMetaDataset(MetaDataset):
    def __init__(
        self,
        info=None,
        rotations=None,
        name="Omniglot",
        num_classes=None,
        num_examples=None,
    ):
        super().__init__(
            info, name=name, num_classes=num_classes, num_examples=num_examples
        )
        self._loaded_images = defaultdict(lambda: {})
        self._rotations = rotations or [0, 90, 180, 270]
        self.load_all()

    def generate_datasets(self, rand=None, num_classes=None, num_examples=None):
        rand = dl.get_rand_state(rand)

        if not num_examples:
            num_examples = self.kwargs["num_examples"]
        if not num_classes:
            num_classes = self.kwargs["num_classes"]

        clss = self._loaded_images if self._loaded_images else self.info["classes"]

        random_classes = rand.choice(
            list(clss.keys()), size=(num_classes,), replace=False
        )
        rand_class_dict = {rnd: k for k, rnd in enumerate(random_classes)}

        _dts = []
        for ns in as_tuple_or_list(num_examples):
            classes = balanced_choice_wr(random_classes, ns, rand)

            all_images = {cls: list(clss[cls]) for cls in classes}
            data, targets, sample_info = [], [], []
            for c in classes:
                rand.shuffle(all_images[c])
                img_name = all_images[c][0]
                all_images[c].remove(img_name)
                sample_info.append({"name": img_name, "label": c})
                data.append(clss[c][img_name])
                targets.append(rand_class_dict[c])

            if self.info["one_hot_enc"]:
                targets = dl.to_one_hot_enc(targets, dimension=num_classes)

            _dts.append(
                dl.Dataset(
                    data=np.array(np.stack(data)),
                    target=targets,
                    sample_info=sample_info,
                    info={"all_classes": random_classes},
                )
            )
        return dl.Datasets.from_list(_dts)

    def load_all(self):
        from imageio import imread
        from scipy.ndimage.interpolation import rotate

        _cls = self.info["classes"]
        _base_folder = self.info["base_folder"]

        for c in _cls:
            all_images = list(_cls[c])
            for img_name in all_images:
                img = imread(join(_base_folder, join(c, img_name)))
                img = 1.0 - np.reshape(img, (28, 28, 1)) / 255.0
                for rot in self._rotations:
                    img = rotate(img, rot, reshape=False)
                    self._loaded_images[c + os.path.sep + "rot_" + str(rot)][
                        img_name
                    ] = img


def meta_omniglot(
    folder=OMNIGLOT_RESIZED,
    std_num_classes=None,
    examples_train=None,
    examples_test=None,
    one_hot_enc=True,
    _rand=0,
    n_splits=None,
):
    """
    Loading function for Omniglot dataset in learning-to-learn version. Use image data as obtained from
    https://github.com/cbfinn/maml/blob/master/data/omniglot_resized/resize_images.py

    :param folder: root folder name.
    :param std_num_classes: standard number of classes for N-way classification
    :param examples_train:standard number of examples to be picked in each generated per classes for training
    (eg .1 shot, examples_train=1)
    :param examples_test: standard number of examples to be picked in each generated per classes for testing
    :param one_hot_enc: one hot encoding
    :param _rand: random seed or RandomState for generate training, validation, testing meta-datasets
                    split
    :param n_splits: num classes per split
    :return: a Datasets of MetaDataset s
    """
    std_num_examples = (
        examples_train * std_num_classes,
        examples_test * std_num_classes,
    )
    alphabets = os.listdir(folder)

    labels_and_images = OrderedDict()
    for alphabet in alphabets:
        base_folder = join(folder, alphabet)
        label_names = os.listdir(base_folder)  # all characters in one alphabet
        labels_and_images.update(
            {
                alphabet + os.path.sep + ln: os.listdir(join(base_folder, ln))
                # all examples of each character
                for ln in label_names
            }
        )

    # divide between training validation and test meta-datasets
    _rand = dl.get_rand_state(_rand)
    all_clss = list(labels_and_images.keys())
    _rand.shuffle(all_clss)
    n_splits = n_splits or (0, 1200, 1300, len(all_clss))

    meta_dts = []
    for start, end in zip(n_splits, n_splits[1:]):
        meta_dts.append(
            OmniglotMetaDataset(
                info={
                    "base_folder": folder,
                    "classes": {k: labels_and_images[k] for k in all_clss[start:end]},
                    "one_hot_enc": one_hot_enc,
                },
                num_classes=std_num_classes,
                num_examples=std_num_examples,
            )
        )

    return dl.Datasets.from_list(meta_dts)


def meta_omniglot_v2(
    folder=OMNIGLOT_RESIZED,
    std_num_classes=None,
    examples_train=None,
    examples_test=None,
    one_hot_enc=True,
    _rand=0,
    n_splits=None,
):
    """
    Loading function for Omniglot dataset in learning-to-learn version. Use image data as obtained from
    https://github.com/cbfinn/maml/blob/master/data/omniglot_resized/resize_images.py

    :param folder: root folder name.
    :param std_num_classes: standard number of classes for N-way classification
    :param examples_train:standard number of examples to be picked in each generated per classes for training
    (eg .1 shot, examples_train=1)
    :param examples_test: standard number of examples to be picked in each generated per classes for testing
    :param one_hot_enc: one hot encoding
    :param _rand: random seed or RandomState for generate training, validation, testing meta-datasets
                    split
    :param n_splits: num classes per split
    :return: a Datasets of MetaDataset s
    """

    class OmniglotMetaDataset(MetaDataset):
        def __init__(
            self,
            info=None,
            rotations=None,
            name="Omniglot",
            num_classes=None,
            num_examples=None,
        ):
            super().__init__(
                info, name=name, num_classes=num_classes, num_examples=num_examples
            )
            self._loaded_images = defaultdict(lambda: {})
            self._rotations = rotations or [0, 90, 180, 270]
            self._img_array = None
            self.load_all()

        def generate_datasets(self, rand=None, num_classes=None, num_examples=None):
            rand = dl.get_rand_state(rand)

            if not num_examples:
                num_examples = self.kwargs["num_examples"]
            if not num_classes:
                num_classes = self.kwargs["num_classes"]

            clss = self._loaded_images if self._loaded_images else self.info["classes"]

            random_classes = rand.choice(
                list(clss.keys()), size=(num_classes,), replace=False
            )
            rand_class_dict = {rnd: k for k, rnd in enumerate(random_classes)}

            _dts = []
            for ns in dl.as_tuple_or_list(num_examples):
                classes = balanced_choice_wr(random_classes, ns, rand)

                all_images = {cls: list(clss[cls]) for cls in classes}
                indices, targets = [], []
                for c in classes:
                    rand.shuffle(all_images[c])
                    img_name = all_images[c][0]
                    all_images[c].remove(img_name)
                    # sample_info.append({'name': img_name, 'label': c})
                    indices.append(clss[c][img_name])
                    targets.append(rand_class_dict[c])

                if self.info["one_hot_enc"]:
                    targets = dl.to_one_hot_enc(targets, dimension=num_classes)

                data = self._img_array[indices]

                _dts.append(dl.Dataset(data=data, target=targets))
            return dl.Datasets.from_list(_dts)

        def load_all(self):
            from scipy.ndimage import imread
            from scipy.ndimage.interpolation import rotate

            _cls = self.info["classes"]
            _base_folder = self.info["base_folder"]

            _id = 0
            flat_data = []
            flat_targets = []
            for c in _cls:
                all_images = list(_cls[c])
                for img_name in all_images:
                    img = imread(join(_base_folder, join(c, img_name)))
                    img = 1.0 - np.reshape(img, (28, 28, 1)) / 255.0
                    for rot in self._rotations:
                        img = rotate(img, rot, reshape=False)
                        self._loaded_images[c + os.path.sep + "rot_" + str(rot)][
                            img_name
                        ] = _id
                        _id += 1
                        flat_data.append(img)
                        # flat_targets maybe... no flat targets... they depend on the episode!!

            self._img_array = np.stack(flat_data)

            # end of class

    std_num_examples = (
        examples_train * std_num_classes,
        examples_test * std_num_classes,
    )
    alphabets = os.listdir(folder)

    labels_and_images = OrderedDict()
    for alphabet in alphabets:
        base_folder = join(folder, alphabet)
        label_names = os.listdir(base_folder)  # all characters in one alphabet
        labels_and_images.update(
            {
                alphabet + os.path.sep + ln: os.listdir(join(base_folder, ln))
                # all examples of each character
                for ln in label_names
            }
        )

    # divide between training validation and test meta-datasets
    _rand = dl.get_rand_state(_rand)
    all_clss = list(labels_and_images.keys())
    _rand.shuffle(all_clss)
    n_splits = n_splits or (0, 1200, 1300, len(all_clss))

    meta_dts = []
    for start, end in zip(n_splits, n_splits[1:]):
        meta_dts.append(
            OmniglotMetaDataset(
                info={
                    "base_folder": folder,
                    "classes": {k: labels_and_images[k] for k in all_clss[start:end]},
                    "one_hot_enc": one_hot_enc,
                },
                num_classes=std_num_classes,
                num_examples=std_num_examples,
            )
        )

    return dl.Datasets.from_list(meta_dts)


class ImageNetMetaDataset(MetaDataset):
    def __init__(
        self,
        info=None,
        name="MiniImagenet",
        num_classes=None,
        num_examples=None,
        h5=False,
    ):

        super().__init__(
            info, name=name, num_classes=num_classes, num_examples=num_examples
        )
        self._loaded_images = defaultdict(lambda: {})
        self._threads = []
        self.h5 = h5

    def load_all_images(self):
        if self.h5:
            _file = self.info["file"]
            h5m = h5py.File(_file, "r")
            img_per_class = 600
            for j in range(len(h5m["X"])):
                self._loaded_images[j // img_per_class][j] = (
                    np.array(h5m["X"][j], dtype=np.float32) / 255.0
                )
                # images were stored as int
        else:
            from imageio import imread

            # from scipy.misc import imresize
            _cls = self.info["classes"]
            _base_folder = self.info["base_folder"]

            def _load_class(c):
                all_images = list(_cls[c])
                for img_name in all_images:
                    img = imread(join(_base_folder, join(c, img_name)))
                    if self.info["resize"]:
                        # noinspection PyTypeChecker
                        # img = imresize(img, size=(self.info['resize'], self.info['resize'], 3)) / 255.
                        img = (
                            np.array(
                                Image.fromarray(img).resize(
                                    size=(self.info["resize"], self.info["resize"])
                                )
                            )
                            / 255.0
                        )
                    self._loaded_images[c][img_name] = img

            for cls in _cls:
                self._threads.append(
                    threading.Thread(target=lambda: _load_class(cls), daemon=True)
                )
                self._threads[-1].start()

    def check_loaded_images(self, n_min):
        return self._loaded_images and all(
            [len(v) >= n_min for v in self._loaded_images.values()]
        )

    def generate_datasets(
        self, rand=None, num_classes=None, num_examples=None, wait_for_n_min=None
    ):

        rand = dl.get_rand_state(rand)

        if wait_for_n_min:
            import time

            while not self.check_loaded_images(wait_for_n_min):
                time.sleep(5)

        if not num_examples:
            num_examples = self.kwargs["num_examples"]
        if not num_classes:
            num_classes = self.kwargs["num_classes"]

        clss = self._loaded_images if self._loaded_images else self.info["classes"]

        random_classes = rand.choice(
            list(clss.keys()), size=(num_classes,), replace=False
        )
        rand_class_dict = {rnd: k for k, rnd in enumerate(random_classes)}

        _dts = []
        for ns in dl.as_tuple_or_list(num_examples):
            classes = balanced_choice_wr(random_classes, ns, rand)

            all_images = {cls: list(clss[cls]) for cls in classes}
            data, targets, sample_info = [], [], []
            for c in classes:
                rand.shuffle(all_images[c])
                img_name = all_images[c][0]
                all_images[c].remove(img_name)
                sample_info.append({"name": img_name, "label": c})

                if self._loaded_images:
                    data.append(clss[c][img_name])
                else:
                    from imageio import imread

                    data.append(
                        np.array(
                            Image.fromarray(
                                imread(
                                    join(self.info["base_folder"], join(c, img_name))
                                )
                            ).resize(size=(self.info["resize"], self.info["resize"]))
                        )
                        / 255.0
                    )
                targets.append(rand_class_dict[c])

            if self.info["one_hot_enc"]:
                targets = to_one_hot_enc(targets, dimension=num_classes)

            _dts.append(
                dl.Dataset(
                    data=np.array(np.stack(data)),
                    target=targets,
                    sample_info=sample_info,
                    info={"all_classes": random_classes},
                )
            )
        return dl.Datasets.from_list(_dts)

    def all_data(self, partition_proportions=None, seed=None):
        if not self._loaded_images:
            self.load_all_images()
            while not self.check_loaded_images(600):
                import time

                time.sleep(5)
        data, targets = [], []
        for k, c in enumerate(sorted(self._loaded_images)):
            data += list(self._loaded_images[c].values())
            targets += [k] * 600
        if self.info["one_hot_enc"]:
            targets = dl.to_one_hot_enc(targets, dimension=len(self._loaded_images))
        _dts = [
            dl.Dataset(
                data=np.stack(data), target=np.array(targets), name="MiniImagenet_full"
            )
        ]
        if seed:
            np.random.seed(seed)
        if partition_proportions:
            _dts = redivide_data(
                _dts, partition_proportions=partition_proportions, shuffle=True
            )
        return dl.Datasets.from_list(_dts)


def meta_mini_imagenet(
    folder=MINI_IMAGENET_FOLDER_V3,
    sub_folders=None,
    std_num_classes=None,
    examples_train=None,
    examples_test=None,
    resize=84,
    one_hot_enc=True,
    load_all_images=True,
    h5=False,
):
    """
    Load a meta-datasets from Mini-ImageNet. Returns a Datasets of MetaDataset s,

    :param folder: base folder
    :param sub_folders: optional sub-folders in which data is locate
    :param std_num_classes: standard number of classes to be included in each generated per dataset
                            (can be None, default)
    :param examples_train:standard number of examples to be picked in each generated per classes for training
    :param examples_test: standard number of examples to be picked in each generated per classes for testing
    :param resize:  resizing dimension
    :param one_hot_enc:
    :param load_all_images:
    :param h5:  True (default) to use HDF5 files, when False search for JPEG images.
    :return:
    """
    std_num_examples = (
        examples_train * std_num_classes,
        examples_test * std_num_classes,
    )

    if sub_folders is None:
        sub_folders = ["train", "val", "test"]
    meta_dts = []
    for ds in sub_folders:
        if not h5:
            base_folder = join(folder, ds)
            label_names = os.listdir(base_folder)
            labels_and_images = {
                ln: os.listdir(join(base_folder, ln)) for ln in label_names
            }
            meta_dts.append(
                ImageNetMetaDataset(
                    info={
                        "base_folder": base_folder,
                        "classes": labels_and_images,
                        "resize": resize,
                        "one_hot_enc": one_hot_enc,
                    },
                    num_classes=std_num_classes,
                    num_examples=std_num_examples,
                    h5=False,
                )
            )
        else:
            file = join(folder, ds + ".h5")
            meta_dts.append(
                ImageNetMetaDataset(
                    info={"file": file, "one_hot_enc": one_hot_enc},
                    num_classes=std_num_classes,
                    num_examples=std_num_examples,
                    h5=True,
                )
            )

    dts = dl.Datasets.from_list(meta_dts)
    if load_all_images:
        import time

        [_d.load_all_images() for _d in dts]
        _check_available = lambda min_num: [
            _d.check_loaded_images(min_num) for _d in dts
        ]
        while not all(_check_available(15)):
            time.sleep(
                1
            )  # be sure that there are at least 15 images per class in each meta-dataset
    return dts


def random_classification_datasets(
    n_samples,
    features=100,
    classes=2,
    informative=0.1,
    partition_proportions=(0.5, 0.3),
    rnd=None,
    one_hot=True,
    **mk_cls_kwargs
):
    rnd_state = dl.get_rand_state(rnd)
    X, Y = make_classification(
        n_samples, features, n_classes=classes, random_state=rnd_state, **mk_cls_kwargs
    )
    if one_hot:
        Y = to_one_hot_enc(Y)

    print("range of Y", np.min(Y), np.max(Y))
    info = merge_dicts({"informative": informative, "random_seed": rnd}, mk_cls_kwargs)
    name = dl.em_utils.name_from_dict(info, "w")
    dt = dl.Dataset(X, Y, name=name, info=info)
    datasets = dl.Datasets.from_list(redivide_data([dt], partition_proportions))
    print(
        "conditioning of X^T X",
        np.linalg.cond(datasets.train.data.T @ datasets.train.data),
    )
    return datasets


def random_regression_datasets(
    n_samples,
    features=100,
    outs=1,
    informative=0.1,
    partition_proportions=(0.5, 0.3),
    rnd=None,
    **mk_rgr_kwargs
):
    rnd_state = dl.get_rand_state(rnd)
    X, Y, w = make_regression(
        n_samples,
        features,
        int(features * informative),
        outs,
        random_state=rnd_state,
        coef=True,
        **mk_rgr_kwargs
    )
    if outs == 1:
        Y = np.reshape(Y, (n_samples, 1))

    print("range of Y", np.min(Y), np.max(Y))
    info = merge_dicts(
        {"informative": informative, "random_seed": rnd, "w": w}, mk_rgr_kwargs
    )
    name = dl.em_utils.name_from_dict(info, "w")
    dt = dl.Dataset(X, Y, name=name, info=info)
    datasets = dl.Datasets.from_list(redivide_data([dt], partition_proportions))
    print(
        "conditioning of X^T X",
        np.linalg.cond(datasets.train.data.T @ datasets.train.data),
    )
    return datasets


def redivide_data(
    datasets,
    partition_proportions=None,
    shuffle=False,
    filters=None,
    maps=None,
    balance_classes=False,
    rand=None,
):
    """
    Function that redivides datasets. Can be use also to shuffle or filter or map examples.

    :param rand:
    :param balance_classes: # TODO RICCARDO
    :param datasets: original datasets, instances of class Dataset (works with get_data and get_targets for
                        compatibility with mnist datasets
    :param partition_proportions: (optional, default None)  list of fractions that can either sum up to 1 or less
                                    then one, in which case one additional partition is created with
                                    proportion 1 - sum(partition proportions).
                                    If None it will retain the same proportion of samples found in datasets
    :param shuffle: (optional, default False) if True shuffles the examples
    :param filters: (optional, default None) filter or list of filters: functions with signature
                        (data, target, index) -> boolean (accept or reject the sample)
    :param maps: (optional, default None) map or list of maps: functions with signature
                        (data, target, index) ->  (new_data, new_target) (maps the old sample to a new one,
                        possibly also to more
                        than one sample, for data augmentation)
    :return: a list of datasets of length equal to the (possibly augmented) partition_proportion
    """

    rnd = get_rand_state(rand)

    all_data = vstack([get_data(d) for d in datasets])
    all_labels = stack_or_concat([get_targets(d) for d in datasets])

    all_infos = np.concatenate([d.sample_info for d in datasets])

    N = all_data.shape[0]

    if partition_proportions:  # argument check
        partition_proportions = list(
            [partition_proportions]
            if isinstance(partition_proportions, float)
            else partition_proportions
        )
        sum_proportions = sum(partition_proportions)
        assert sum_proportions <= 1, (
            "partition proportions must sum up to at most one: %d" % sum_proportions
        )
        if sum_proportions < 1.0:
            partition_proportions += [1.0 - sum_proportions]
    else:
        partition_proportions = [1.0 * get_data(d).shape[0] / N for d in datasets]

    if shuffle:
        if sp and isinstance(all_data, sp.sparse.csr.csr_matrix):
            raise NotImplementedError()
        # if sk_shuffle:  # TODO this does not work!!! find a way to shuffle these matrices while
        # keeping compatibility with tensorflow!
        #     all_data, all_labels, all_infos = sk_shuffle(all_data, all_labels, all_infos)
        # else:
        permutation = np.arange(all_data.shape[0])
        rnd.shuffle(permutation)

        all_data = all_data[permutation]
        all_labels = np.array(all_labels[permutation])
        all_infos = np.array(all_infos[permutation])

    if filters:
        if sp and isinstance(all_data, sp.sparse.csr.csr_matrix):
            raise NotImplementedError()
        filters = as_list(filters)
        data_triple = [(x, y, d) for x, y, d in zip(all_data, all_labels, all_infos)]
        for fiat in filters:
            data_triple = [
                xy for i, xy in enumerate(data_triple) if fiat(xy[0], xy[1], xy[2], i)
            ]
        all_data = np.vstack([e[0] for e in data_triple])
        all_labels = np.vstack([e[1] for e in data_triple])
        all_infos = np.vstack([e[2] for e in data_triple])

    if maps:
        if sp and isinstance(all_data, sp.sparse.csr.csr_matrix):
            raise NotImplementedError()
        maps = as_list(maps)
        data_triple = [(x, y, d) for x, y, d in zip(all_data, all_labels, all_infos)]
        for _map in maps:
            data_triple = [
                _map(xy[0], xy[1], xy[2], i) for i, xy in enumerate(data_triple)
            ]
        all_data = np.vstack([e[0] for e in data_triple])
        all_labels = np.vstack([e[1] for e in data_triple])
        all_infos = np.vstack([e[2] for e in data_triple])

    N = all_data.shape[0]
    assert N == all_labels.shape[0]

    calculated_partitions = reduce(
        lambda v1, v2: v1 + [sum(v1) + v2],
        [int(N * prp) for prp in partition_proportions],
        [0],
    )
    calculated_partitions[-1] = N

    print(
        "datasets.redivide_data:, computed partitions numbers -",
        calculated_partitions,
        "len all",
        N,
        end=" ",
    )

    new_general_info_dict = {}
    for data in datasets:
        new_general_info_dict = {**new_general_info_dict, **data.info}

        if balance_classes:
            new_datasets = []
            forbidden_indices = np.empty(0, dtype=np.int64)
            for d1, d2 in zip(calculated_partitions[:-1], calculated_partitions[1:-1]):
                indices = np.array(
                    get_indices_balanced_classes(d2 - d1, all_labels, forbidden_indices)
                )
                dataset = dl.Dataset(
                    data=all_data[indices],
                    target=all_labels[indices],
                    sample_info=all_infos[indices],
                    info=new_general_info_dict,
                )
                new_datasets.append(dataset)
                forbidden_indices = np.append(forbidden_indices, indices)
                test_if_balanced(dataset)
            remaining_indices = np.array(
                list(set(list(range(N))) - set(forbidden_indices))
            )
            new_datasets.append(
                dl.Dataset(
                    data=all_data[remaining_indices],
                    target=all_labels[remaining_indices],
                    sample_info=all_infos[remaining_indices],
                    info=new_general_info_dict,
                )
            )
        else:
            new_datasets = [
                dl.Dataset(
                    data=all_data[d1:d2],
                    target=all_labels[d1:d2],
                    sample_info=all_infos[d1:d2],
                    info=new_general_info_dict,
                )
                for d1, d2 in zip(calculated_partitions, calculated_partitions[1:])
            ]

        print("DONE")

        return new_datasets


if __name__ == "__main__":
    # pass
    # mmi = meta_mini_imagenet()
    # # st = mmi.train.generate_datasets(num_classes=10, num_examples=(123, 39))
    # d1 = mmi.train.all_data(seed=0)
    # print(d1.train.dim_data, d1.train.dim_target)
    #
    # mmiii = meta_mini_imagenet()
    # # st = mmi.train.generate_datasets(num_classes=10, num_examples=(123, 39))
    # d2 = mmiii.train.all_data(seed=0)
    # print(d2.train.dim_data, d2.train.dim_target)
    #
    # print(np.equal(d1.train.data[0], d2.train.data[0]))

    res = meta_omniglot_v2(std_num_classes=5, std_num_examples=(10, 20))
    lst = []
    while True:
        lst.append(res.train.generate_datasets())
        # print(dt.train.data)
        # print(res)



from boml.load_data.datasets.dl_utils import *
from boml.load_data.datasets.structures import Dataset, MetaDataset
from boml.load_data.datasets.load_full_dataset import (
    meta_omniglot,
    meta_mini_imagenet,
    OmniglotMetaDataset,
    ImageNetMetaDataset,
    mnist,
)



from collections import OrderedDict

import tensorflow as tf

from boml.lower_iter import BOMLInnerGradTrad


class BOMLInnerGradAggr(BOMLInnerGradTrad):
    def __init__(self, update_op, dynamics, objective, outer_objective):
        """
        :param update_op: the tf operation to perform updates
        :param dynamics: the iterative formats of dynamical system
        :param objective: Lower-Level objective
        :param outer_objective: Upper-Level objective
        """
        self.outer_objective = outer_objective
        super().__init__(update_op=update_op, dynamics=dynamics, objective=objective)

    @staticmethod
    def compute_gradients(
        boml_opt,
        loss_inner,
        loss_outer=None,
        param_dict=OrderedDict(),
        var_list=None,
        **inner_kargs
    ):
        """
        :param boml_opt: instance of modified optimizers in the `optimizer` module
        :param loss_inner: Lower-Level objectives
        :param loss_outer: Upper-Level objectives
        :param param_dict: dictionary of general parameters for different algorithms
        :param var_list: the list of parameters in the base-learner
        :param inner_kargs: optional arguments for tensorflow optimizers, like global_step, gate_gradients
        :return: initialized instance of inner_grad for UL optimization
        """
        minimize_kargs = {
            inner_arg: inner_kargs[inner_arg]
            for inner_arg in set(inner_kargs.keys()) - set(param_dict.keys())
        }

        assert loss_inner is not None, "argument:inner_objective must be initialized"
        assert {
            "alpha",
            "s",
            "t",
            "t_tensor",
        } <= param_dict.keys(), (
            "Necessary hyper_parameters must be initialized before calling minimize()"
        )
        # alpha, loss_outer, s, t, t_tensor = sorted(param_dict.items(), key=lambda x: x[0])
        update_op, dynamics = BOMLInnerGradAggr.bml_inner_grad_aggr(
            inner_optimizer=boml_opt,
            loss_inner=loss_inner,
            loss_outer=loss_outer,
            param_dict=param_dict,
            var_list=var_list,
            *minimize_kargs
        )
        return BOMLInnerGradAggr(
            update_op=update_op,
            dynamics=dynamics,
            objective=loss_inner,
            outer_objective=loss_outer,
        )

    @staticmethod
    def bml_inner_grad_aggr(
        inner_optimizer,
        loss_inner,
        loss_outer,
        param_dict=OrderedDict(),
        global_step=None,
        var_list=None,
        gate_gradients=1,
        aggregation_method=None,
        colocate_gradients_with_ops=False,
        name=None,
        grad_loss=None,
    ):

        grads_and_vars_inner = inner_optimizer.compute_gradients(
            loss_inner,
            var_list=var_list,
            gate_gradients=gate_gradients,
            aggregation_method=aggregation_method,
            colocate_gradients_with_ops=colocate_gradients_with_ops,
            grad_loss=grad_loss,
        )
        grads_and_vars_outer = inner_optimizer.compute_gradients(
            loss_outer,
            var_list=var_list,
            gate_gradients=gate_gradients,
            aggregation_method=aggregation_method,
            colocate_gradients_with_ops=colocate_gradients_with_ops,
            grad_loss=grad_loss,
        )
        grads_and_vars = BOMLInnerGradAggr.combine_grads(
            inner_grads=grads_and_vars_inner,
            outer_grads=grads_and_vars_outer,
            alpha=param_dict["alpha"],
            s=param_dict["s"],
            t=param_dict["t"],
            t_tensor=param_dict["t_tensor"],
        )

        return inner_optimizer.apply_gradients(
            grads_and_vars, global_step=global_step, name=name
        )

    @staticmethod
    def combine_grads(inner_grads, outer_grads, alpha, s, t, t_tensor):

        combine_grads = []
        if not alpha.get_shape().as_list():
            for _ in range(len(inner_grads)):
                ll_part = (1 - alpha / t_tensor) * t * inner_grads[_][0]
                ul_part = (alpha / t_tensor) * s * outer_grads[_][0]
                combine_grads.append((ll_part + ul_part, inner_grads[_][1]))
        else:
            for _ in range(len(inner_grads)):
                ll_part = (
                    (1 - tf.norm(tf.matmul(alpha, t_tensor), ord=1))
                    * t
                    * inner_grads[_][0]
                )
                ul_part = (
                    tf.norm(tf.matmul(alpha, t_tensor), ord=1) * s * outer_grads[_][0]
                )
                combine_grads.append((ll_part + ul_part, inner_grads[_][1]))
        return combine_grads



from __future__ import absolute_import, print_function, division

from collections import OrderedDict

import tensorflow as tf
from tensorflow.contrib.opt import ScipyOptimizerInterface
from tensorflow.python.training import slot_creator

from boml import utils
from boml.upper_iter import BOMLOuterGrad

RAISE_ERROR_ON_DETACHED = False


class BOMLOuterGradImplicit(BOMLOuterGrad):
    """
    Implementation follows Pedregosa's algorithm HOAG
    """

    def __init__(
        self,
        inner_method="Trad",
        linear_system_solver_gen=None,
        name="BMLOuterGradImplicit",
    ):
        super(BOMLOuterGradImplicit, self).__init__(name)
        self._inner_method = inner_method
        if linear_system_solver_gen is None:
            linear_system_solver_gen = lambda _obj, var_list, _tolerance: ScipyOptimizerInterface(
                _obj,
                var_list=var_list,
                options={"maxiter": 5},
                method="cg",
                tol=_tolerance,
            )
        self.linear_system_solver = linear_system_solver_gen

        self.tolerance = lambda _k: 0.1 * (0.9 ** _k)

        self._lin_sys = []
        self._qs = []

    def compute_gradients(
        self, outer_objective, inner_grad, meta_param=None, param_dict=OrderedDict()
    ):
        meta_param = super(BOMLOuterGradImplicit, self).compute_gradients(
            outer_objective, inner_grad, meta_param
        )
        state = list(inner_grad.state)

        with tf.variable_scope(outer_objective.op.name):
            g1 = utils.vectorize_all(
                tf.gradients(outer_objective, state)
            )  # Lower Level gradient of UL objective  w.r.t task parameters
            grads_inner_obj_vec = utils.vectorize_all(
                tf.gradients(inner_grad.objective, state)
            )  #  Lower Level gradient of LL objective  w.r.t task parameters

            q = self._create_q(g1)
            obj = tf.norm(
                utils.vectorize_all(
                    tf.gradients(utils.dot(grads_inner_obj_vec, q), state)
                )
                - g1
            )  # using the norm seems to produce better results then squared norm...
            # (even though is more costly)

            self._lin_sys.append(
                lambda _tolerance: self.linear_system_solver(obj, [q], _tolerance)
            )

            g2s = tf.gradients(outer_objective, meta_param)
            cross_ders = tf.gradients(utils.dot(grads_inner_obj_vec, q), meta_param)
            for g2, cd, hyper in zip(g2s, cross_ders, meta_param):
                assert (
                    g2 is not None or cd is not None
                ), BOMLOuterGrad._ERROR_HYPER_DETACHED.format(hyper)
                hg = utils.maybe_add(-cd, g2)
                if hg is None:  # this would be strange...
                    print(
                        "WARNING, outer objective is only directly dependent on hyperparameter {}. "
                        + "Direct optimization would be better!".format(hyper)
                    )
                    hg = g2
                self._hypergrad_dictionary[hyper].append(hg)

            return meta_param

    def _create_q(self, d_oo_d_state):
        self._qs.append(slot_creator.create_zeros_slot(d_oo_d_state, "q"))
        return self._qs[-1]

    def apply_gradients(
        self,
        inner_objective_feed_dicts=None,
        outer_objective_feed_dicts=None,
        initializer_feed_dict=None,
        param_dict=OrderedDict(),
        global_step=None,
        session=None,
    ):
        ss = session or tf.get_default_session()

        inner_objective_feed_dicts = utils.as_tuple_or_list(inner_objective_feed_dicts)

        self._run_batch_initialization(
            ss,
            utils.maybe_call(initializer_feed_dict, utils.maybe_eval(global_step, ss)),
        )

        for t in utils.solve_int_or_generator(param_dict["T"]):
            _fd = utils.maybe_call(inner_objective_feed_dicts[0], t)
            self._forward_step(ss, _fd)

        # end of optimization. Solve linear systems.
        tol_val = utils.maybe_call(
            self.tolerance, utils.maybe_eval(global_step, ss)
        )  # decreasing tolerance (seq.)
        # feed dictionaries (could...in theory, implement stochastic solution of this linear system...)
        _fd = utils.maybe_call(inner_objective_feed_dicts[-1], -1)
        _fd_outer = utils.maybe_call(
            outer_objective_feed_dicts, utils.maybe_eval(global_step, ss)
        )
        _fd = utils.merge_dicts(_fd, _fd_outer)

        for lin_sys in self._lin_sys:
            lin_sys(tol_val).minimize(
                ss, _fd
            )  # implicitly warm restarts with previously found q

    def _forward_step(self, ss, _fd):
        ss.run(self.iteration, _fd)

    def _run_batch_initialization(self, ss, fd):
        ss.run(self.initialization, feed_dict=fd)

    def set_tolerance(self, tolerance=lambda _k: 0.1 * (0.9 ** _k)):
        self.tolerance = tolerance



from __future__ import absolute_import, print_function, division

import sys

import numpy as np
import tensorflow as tf

# noinspection PyClassHasNoInit
import boml

Meta_Init_Method = ["MAML", "FOMAML", "WarpGrad", "MT-net"]
Meta_Init_Method = ["Reverse", "Truncated", "DARTS", "BA", "Implicit"]
METHOD_COLLECTIONS = [Meta_Init_Method, Meta_Init_Method]


def remove_from_collection(key, *lst):
    """
    Remove tensors in lst from collection given by key
    """
    try:
        # noinspection PyProtectedMember
        [tf.get_default_graph()._collections[key].remove(_e) for _e in lst]
    except ValueError:
        print(
            "WARNING: Collection -> {} <- does not contain some tensor in {}".format(
                key, lst
            ),
            file=sys.stderr,
        )


def as_tuple_or_list(obj):
    """
    Make sure that `obj` is a tuple or a list and eventually converts it into a list with a single element

    :param obj:
    :return: A `tuple` or a `list`
    """
    return obj if isinstance(obj, (list, tuple)) else [obj]


def merge_dicts(*dicts):
    """
    Merges dictionaries recursively. Accepts also `None` and returns always a (possibly empty) dictionary
    """
    from functools import reduce

    # if len(dicts) == 1: return dicts[0]
    return reduce(lambda a, nd: merge_two_dicts(a, nd if nd else {}), dicts, {})


def merge_two_dicts(x, y):
    z = x.copy()  # start with x's keys and values
    z.update(y)  # modifies z with y's keys and values & returns None
    return z


def as_list(obj):
    """
    Makes sure `obj` is a list or otherwise converts it to a list with a single element.
    """
    return obj if isinstance(obj, list) else [obj]


def vectorize_all(var_list, name=None):
    """Given a list of tensors returns their concatenated vectorization.
    Note that for matrices the vectorization is row-wise instead of column-wise as
    it should be in Magnus. Could it be a problem?

    :param var_list: **bold**
    :param name: optional name for resulting tensor

    :return: vectorization of `var_list`"""
    with tf.name_scope(name, "Vectorization", var_list) as scope:
        return tf.concat([tf.reshape(_w, [-1]) for _w in var_list], 0, name=scope)


def add_list(lst1, lst2):
    """
    sum elements in two lists by sequence
    :return: new list that sum elements in two lists by sequence
    """
    assert len(lst1) == len(lst2), "length of two list must be equal"
    return [ls1 + ls2 for ls1, ls2 in zip(lst1, lst2)]


def maybe_call(obj, *args, **kwargs):
    """
    Calls obj with args and kwargs and return its result if obj is callable, otherwise returns obj.
    """
    if callable(obj):
        return obj(*args, **kwargs)
    return obj


def dot(a, b, name=None):
    """
    Dot product between vectors `a` and `b` with optional name.
    If a and b are not vectors, formally this computes <vec(a), vec(b)>.
    """
    with tf.name_scope(name, "Dot", [a, b]):
        return tf.reduce_sum(a * b)


def maybe_eval(a, ss=None):
    """
    Run or eval `a` and returns the result if possible.

    :param a: object, or `tf.Variable` or `tf.Tensor`
    :param ss: `tf.Session` or get default session (if any)
    :return: If a is not a tensorflow evaluable returns it, or returns the
                resulting call
    """
    if ss is None:
        ss = tf.get_default_session()
    if hasattr(a, "eval") or hasattr(a, "run"):
        return ss.run(a)
    return a


def solve_int_or_generator(int_or_generator):
    return range(int_or_generator) if isinteger(int_or_generator) else int_or_generator


def cross_entropy(pred, label, method="MetaInit"):

    # Note - with tf version <=0.12, this loss has incorrect 2nd derivatives
    assert method in ("MetaInit", "MetaRepr"), "Wrong value for argument method"
    if method == "MetaInit":
        return tf.nn.softmax_cross_entropy_with_logits(logits=pred, labels=label)
    elif method == "MetaRepr":
        return tf.reduce_mean(
            tf.nn.softmax_cross_entropy_with_logits(logits=pred, labels=label)
        )


def classification_acc(pred, label):
    return tf.contrib.metrics.accuracy(
        tf.argmax(tf.nn.softmax(pred), 1), tf.argmax(label, 1)
    )


def set_gpu():
    # set general configuration
    gpu_config = tf.ConfigProto(allow_soft_placement=True)
    gpu_config.gpu_options.allow_growth = True
    return gpu_config


def get_rand_state(rand):
    """
    Utility methods for getting a `RandomState` object.

    :param rand: rand can be None (new State will be generated),
                    np.random.RandomState (it will be returned) or an integer (will be treated as seed).

    :return: a `RandomState` object
    """
    if isinstance(rand, np.random.RandomState):
        return rand
    elif isinstance(rand, (int, np.ndarray, list)) or rand is None:
        return np.random.RandomState(rand)
    else:
        raise ValueError("parameter rand {} has wrong type".format(rand))


def maybe_add(a, b):
    """
    return a if b is None else a + b
    """
    return a if b is None else a + b


def val_or_zero(a, b):
    """
    return a if a is not None else tf.zeros_like(b)
    """
    return a if a is not None else tf.zeros_like(b)


def isinteger(num):
    return isinstance(num, (int, np.int_, np.int8, np.int16, np.int32, np.int64))


def feed_dicts(dat_lst, exs):
    dat_lst = boml.utils.as_list(dat_lst)
    train_fd = boml.utils.merge_dicts(
        *[
            {_ex.x: dat.train.data, _ex.y: dat.train.target}
            for _ex, dat in zip(exs, dat_lst)
        ]
    )
    valid_fd = boml.utils.merge_dicts(
        *[
            {_ex.x_: dat.test.data, _ex.y_: dat.test.target}
            for _ex, dat in zip(exs, dat_lst)
        ]
    )

    return train_fd, valid_fd



from __future__ import print_function, absolute_import, division

from collections import OrderedDict

import tensorflow as tf

from boml.optimizer.opt import BOMLOpt

GRADIENT_NONE_MESSAGE = (
    "WARNING: the gradient w.r.t.the tf.Variable\n {}\n is None;\n "
    "Check the computational graph of the inner objective, and be sure you\n"
    "are not considering including variables that should not be there among the\n"
    "inner variables."
)


class BOMLOptMomentum(BOMLOpt, tf.train.MomentumOptimizer):
    def __init__(
        self,
        learning_rate,
        momentum,
        use_locking=False,
        name="Momentum",
        use_nesterov=False,
    ):
        assert use_nesterov is False, "Nesterov momentum not implemented yet..."
        super(BOMLOptMomentum, self).__init__(
            learning_rate, momentum, use_locking, name, use_nesterov
        )

    def apply_gradients(self, grads_and_vars, global_step=None, name=None):

        update_op = super(BOMLOptMomentum, self).apply_gradients(
            grads_and_vars, global_step, name
        )

        # builds up the dynamics here
        mn = self.get_slot_names()[0]
        dynamics = OrderedDict()
        for g, w in grads_and_vars:
            assert g is not None, GRADIENT_NONE_MESSAGE.format(w)

            m = self.get_slot(w, mn)
            mk = tf.cast(self._momentum_tensor, m.dtype) * m + g
            wk = w - tf.cast(self._learning_rate_tensor, mk.dtype) * mk
            dynamics[w] = wk
            dynamics[m] = mk

        return update_op, dynamics

    def __str__(self):
        return "{}-lr={}-m={}".format(self._name, self._learning_rate, self._momentum)

    @property
    def optimizer_params_tensor(self):
        return super(BOMLOptMomentum, self).optimizer_params_tensor + [
            self._momentum_tensor
        ]



from boml.load_data.datasets.dl_utils import *
from boml.load_data.datasets import (
    meta_mini_imagenet,
    meta_omniglot,
    OmniglotMetaDataset,
    ImageNetMetaDataset,
    mnist,
)
from boml.load_data.experiment import BOMLExperiment
from boml.load_data.datasets.structures import MetaDataset, Datasets, Dataset



from __future__ import absolute_import, print_function, division

from collections import OrderedDict, deque

# import py_bml.OuterOpt.outer_opt_utils as utils
import numpy as np
import tensorflow as tf
from tensorflow.python.training import slot_creator

import boml.extension
from boml import utils
from boml.upper_iter.outer_grad import BOMLOuterGrad

RAISE_ERROR_ON_DETACHED = False


class BOMLOuterGradReverse(BOMLOuterGrad):
    def __init__(
        self, inner_method="Trad", truncate_iter=-1, name="BMLOuterGradReverse"
    ):
        """
       Utility method to initialize truncated reverse HG (not necessarily online),
       :param truncate_iter: Maximum number of iterations that will be stored
       :param name: a name for the operations and variables that will be created
       :return: ReverseHG object
           """
        super(BOMLOuterGradReverse, self).__init__(name)
        self._inner_method = inner_method
        self._alpha_iter = tf.no_op()
        self._reverse_initializer = tf.no_op()
        self._diff_initializer = tf.no_op()
        self._darts_initializer = tf.no_op()
        self._history = deque(maxlen=truncate_iter + 1) if truncate_iter >= 0 else []

    # noinspection SpellCheckingInspection
    def compute_gradients(
        self, outer_objective, inner_grad, meta_param=None, param_dict=OrderedDict()
    ):
        """
        Function that adds to the computational graph all the operations needend for computing
        the hypergradients in a "dynamic" way, without unrolling the entire optimization graph.
        The resulting computation, while being roughly 2x more expensive then unrolling the
        optimizaiton dynamics, requires much less (GPU) memory and is more flexible, allowing
        to set a termination condition to the parameters optimizaiton routine.

        :param inner_grad: OptimzerDict object resulting from the inner objective optimization.
        :param outer_objective: A loss function for the outer parameters (scalar tensor)
        :param meta_param: Optional list of outer parameters to consider. If not provided will get all variables in the
                            hyperparameter collection in the current scope.

        :return: list of outer parameters involved in the computation
        """
        meta_param = super(BOMLOuterGradReverse, self).compute_gradients(
            outer_objective, inner_grad, meta_param
        )

        with tf.variable_scope(outer_objective.op.name):
            doo_ds = tf.gradients(outer_objective, list(inner_grad.state))
            alphas = self._create_lagrangian_multipliers(inner_grad, doo_ds)

            alpha_vec = utils.vectorize_all(alphas)
            dyn_vec = utils.vectorize_all(list(inner_grad.dynamics))
            lag_phi_t = utils.dot(alpha_vec, dyn_vec, name="iter_wise_lagrangian_part1")

            alpha_dot_B = tf.gradients(lag_phi_t, meta_param)

            hyper_grad_vars, hyper_grad_step = [], tf.no_op()
            for dl_dh, hyper in zip(alpha_dot_B, meta_param):
                assert dl_dh is not None, BOMLOuterGrad._ERROR_HYPER_DETACHED.format(
                    hyper
                )
                hgv = None
                if dl_dh is not None:
                    hgv = self._create_outergradient(outer_objective, hyper)

                    hyper_grad_step = tf.group(hyper_grad_step, hgv.assign_add(dl_dh))
                hyper_grad_vars.append(hgv)
                # first update hypergradinet then alphas.
            with tf.control_dependencies([hyper_grad_step]):
                _alpha_iter = tf.group(
                    *[
                        alpha.assign(dl_ds)
                        for alpha, dl_ds in zip(
                            alphas, tf.gradients(lag_phi_t, list(inner_grad.state))
                        )
                    ]
                )
            self._alpha_iter = tf.group(self._alpha_iter, _alpha_iter)
            # put all the backward iterations toghether
            [
                self._hypergrad_dictionary[h].append(hg)
                for h, hg in zip(meta_param, hyper_grad_vars)
            ]
            self._reverse_initializer = tf.group(
                self._reverse_initializer,
                tf.variables_initializer(alphas),
                tf.variables_initializer(
                    [h for h in hyper_grad_vars if hasattr(h, "initializer")]
                ),
            )
            return meta_param

    @staticmethod
    def _create_lagrangian_multipliers(optimizer_dict, doo_ds):
        lag_mul = [
            slot_creator.create_slot(
                v.initialized_value(), utils.val_or_zero(der, v), "alpha"
            )
            for v, der in zip(optimizer_dict.state, doo_ds)
        ]
        [
            tf.add_to_collection(boml.extension.GraphKeys.LAGRANGIAN_MULTIPLIERS, lm)
            for lm in lag_mul
        ]
        boml.extension.remove_from_collection(
            boml.extension.GraphKeys.GLOBAL_VARIABLES, *lag_mul
        )
        # this prevents the 'automatic' initialization with tf.global_variables_initializer.
        return lag_mul

    @staticmethod
    def _create_outergradient_from_dodh(hyper, doo_dhypers):
        """
        Creates one hyper-gradient as a variable. doo_dhypers:  initialization, that is the derivative of
        the outer objective w.r.t this hyper
        """
        hgs = slot_creator.create_slot(
            hyper, utils.val_or_zero(doo_dhypers, hyper), "outergradient"
        )
        boml.extension.remove_from_collection(
            boml.extension.GraphKeys.GLOBAL_VARIABLES, hgs
        )
        return hgs

    @staticmethod
    def _create_outergradient(outer_obj, hyper):
        return BOMLOuterGradReverse._create_outergradient_from_dodh(
            hyper, tf.gradients(outer_obj, hyper)[0]
        )

    def _state_feed_dict_generator(self, history, T_or_generator):
        for t, his in zip(utils.solve_int_or_generator(T_or_generator), history):
            yield t, utils.merge_dicts(
                *[
                    od.state_feed_dict(h)
                    for od, h in zip(sorted(self._optimizer_dicts), his)
                ]
            )

    def apply_gradients(
        self,
        inner_objective_feed_dicts=None,
        outer_objective_feed_dicts=None,
        initializer_feed_dict=None,
        param_dict=OrderedDict(),
        train_batches=None,
        experiments=[],
        global_step=None,
        session=None,
    ):

        if self._inner_method == "Aggr":
            alpha = param_dict["alpha"]
            t_tensor = param_dict["t_tensor"]

        # same thing for T
        T_or_generator = utils.as_tuple_or_list(param_dict["T"])

        ss = session or tf.get_default_session()

        self._history.clear()

        _fd = utils.maybe_call(initializer_feed_dict, utils.maybe_eval(global_step, ss))
        self._save_history(ss.run(self.initialization, feed_dict=_fd))

        T = 0  # this is useful if T_or_generator is indeed a generator...
        for t in utils.solve_int_or_generator(T_or_generator[0]):
            # nonlocal t  # with nonlocal would not be necessary the variable T... not compatible with 2.7

            _fd = inner_objective_feed_dicts
            if self._inner_method == "Aggr":
                _fd.update(outer_objective_feed_dicts)
                if not alpha.get_shape().as_list():
                    _fd[t_tensor] = float(t + 1.0)
                else:
                    tmp = np.zeros((alpha.get_shape().as_list()[1], 1))
                    tmp[t][0] = 1.0
                    _fd[t_tensor] = tmp

            self._save_history(ss.run(self.iteration, feed_dict=_fd))
            T = t

        # initialization of support variables (supports stochastic evaluation of outer objective via global_step ->
        # variable)
        reverse_init_fd = utils.maybe_call(
            outer_objective_feed_dicts, utils.maybe_eval(global_step, ss)
        )
        # now adding also the initializer_feed_dict because of tf quirk...
        maybe_init_fd = utils.maybe_call(
            initializer_feed_dict, utils.maybe_eval(global_step, ss)
        )
        reverse_init_fd = utils.merge_dicts(reverse_init_fd, maybe_init_fd)
        ss.run(self._reverse_initializer, feed_dict=reverse_init_fd)

        del self._history[-1]  # do not consider last point

        for pt, state_feed_dict in self._state_feed_dict_generator(
            reversed(self._history), T_or_generator[-1]
        ):
            # this should be fine also for truncated reverse... but check again the index t
            t = (
                T - pt - 1
            )  # if T is int then len(self.history) is T + 1 and this numerator

            new_fd = utils.merge_dicts(state_feed_dict, inner_objective_feed_dicts)

            if self._inner_method == "Aggr":
                new_fd = utils.merge_dicts(new_fd, outer_objective_feed_dicts)
                # modified - mark
                if not alpha.shape.as_list():
                    new_fd[t_tensor] = float(t + 2.0)
                else:
                    tmp = np.zeros((alpha.get_shape().as_list()[1], 1))
                    tmp[t][0] = 1
                    new_fd[t_tensor] = tmp
            ss.run(self._alpha_iter, new_fd)

    def _save_history(self, weights):
        self._history.append(weights)



from __future__ import print_function, absolute_import, division

from collections import OrderedDict

import tensorflow as tf

from boml.optimizer.opt import BOMLOpt

GRADIENT_NONE_MESSAGE = (
    "WARNING: the gradient w.r.t.the tf.Variable\n {}\n is None;\n "
    "Check the computational graph of the inner objective, and be sure you\n"
    "are not considering including variables that should not be there among the\n"
    "inner variables."
)


class BOMLOptAdam(BOMLOpt, tf.train.AdamOptimizer):
    def __init__(
        self,
        learning_rate=0.001,
        beta1=0.9,
        beta2=0.999,
        epsilon=1e-8,
        use_locking=False,
        name="Adam",
    ):
        super(BOMLOptAdam, self).__init__(
            learning_rate, beta1, beta2, epsilon, use_locking, name
        )

    def apply_gradients(self, grads_and_vars, global_step=None, name=None):
        update_op = super(BOMLOptAdam, self).apply_gradients(
            grads_and_vars, global_step, name
        )

        mn, vn = self.get_slot_names()
        dynamics = OrderedDict()

        with tf.name_scope(name, "Adam_Dynamics"):
            try:
                b1_pow, b2_pow = self._beta1_power, self._beta2_power
            except AttributeError:
                b1_pow, b2_pow = self._get_beta_accumulators()
            lr_k = self._lr_t * tf.sqrt(1.0 - b2_pow) / (1.0 - b1_pow)

            lr_k = tf.cast(lr_k, grads_and_vars[0][0].dtype)
            self._beta1_t = tf.cast(self._beta1_t, grads_and_vars[0][0].dtype)
            self._beta2_t = tf.cast(self._beta2_t, grads_and_vars[0][0].dtype)
            self._epsilon_t = tf.cast(self._epsilon_t, grads_and_vars[0][0].dtype)

            for g, w in grads_and_vars:
                assert g is not None, GRADIENT_NONE_MESSAGE.format(w)

                m = self.get_slot(w, mn)
                v = self.get_slot(w, vn)
                mk = tf.add(
                    self._beta1_t * m, (1.0 - self._beta1_t) * g, name=m.op.name
                )
                vk = tf.add(
                    self._beta2_t * v, (1.0 - self._beta2_t) * g * g, name=v.op.name
                )

                wk = tf.subtract(
                    w, lr_k * mk / (tf.sqrt(vk + self._epsilon_t ** 2)), name=w.op.name
                )
                # IMPORTANT NOTE: epsilon should be outside sqrt as from the original implementation,
                # but this brings to numerical instability of the outergradient.

                dynamics[w] = wk
                dynamics[m] = mk
                dynamics[v] = vk

            b1_powk = b1_pow * self._beta1_t
            b2_powk = b2_pow * self._beta2_t

            dynamics[b1_pow] = b1_powk
            dynamics[b2_pow] = b2_powk
        return update_op, dynamics

    def __str__(self):
        return "{}-lr={}-b1={}-b=2{}-ep={}".format(
            self._name, self._lr, self._beta1, self._beta2, self._epsilon
        )

    @property
    def learning_rate(self):
        return self._lr

    @property
    def learning_rate_tensor(self):
        return self._lr_t

    @property
    def optimizer_params_tensor(self):
        return super(BOMLOptAdam, self).optimizer_params_tensor + [
            self._beta1_t,
            self._beta2_t,
        ]



from __future__ import absolute_import, print_function, division

from collections import OrderedDict

import numpy as np

# import py_bml.OuterOpt.outer_opt_utils as utils
import tensorflow as tf
from tensorflow.python.training import slot_creator

import boml.extension
from boml import utils
from boml.upper_iter.outer_grad import BOMLOuterGrad

RAISE_ERROR_ON_DETACHED = False


class BOMLOuterGradDarts(BOMLOuterGrad):
    def __init__(self, inner_method="Trad", name="BMLOuterGradDarts"):
        """
       Utility method to initialize truncated reverse HG (not necessarily online),
       :param name: a name for the operations and variables that will be created
       :return: BMLOuterGradDarts object
           """
        super(BOMLOuterGradDarts, self).__init__(name)
        self.epsilon = 0.0
        self._inner_method = inner_method
        self._diff_initializer = tf.no_op()
        self._darts_initializer = tf.no_op()
        self._history = []

    # noinspection SpellCheckingInspection
    def compute_gradients(
        self, outer_objective, inner_grad, meta_param=None, param_dict=OrderedDict()
    ):
        """
        Function that adds to the computational graph all the operations needend for computing
        the hypergradients in a "dynamic" way, without unrolling the entire optimization graph.
        The resulting computation, while being roughly 2x more expensive then unrolling the
        optimizaiton dynamics, requires much less (GPU) memory and is more flexible, allowing
        to set a termination condition to the parameters optimizaiton routine.

        :param inner_grad: BOMLInnerGrad object resulting from the LL objective optimization.
        :param outer_objective: A loss function for the outer parameters
        :param meta_param: Optional list of outer parameters to consider. If not provided will get all variables in the
                            hyperparameter collection in the current scope.
        :param param_dict: dictionary to store necessary parameters

        :return: list of outer parameters involved in the computation
        """
        meta_param = super(BOMLOuterGradDarts, self).compute_gradients(
            outer_objective, inner_grad, meta_param
        )

        with tf.variable_scope(outer_objective.op.name):

            ex = self.param_dict["experiment"]
            model = self.param_dict["experiment"].model
            loss_func = self.param_dict["loss_func"]

            # compute the first-order gradient of updated outer parameters with ones-step forward
            grads_outer = [
                self._create_outergradient(outer_objective, hyper)
                for hyper in meta_param
            ]

            # compute the first-order gradient of  the initial task parameters
            darts_derivatives = [
                grad for grad in tf.gradients(outer_objective, list(inner_grad.state))
            ]

            # compute the differentiation part, multiplied by Epsilon
            darts_vector = tf.concat(
                axis=0, values=utils.vectorize_all(darts_derivatives)
            )
            self.epsilon = 0.01 / tf.norm(tensor=darts_vector, ord=2)
            darts_derivaives = [
                self.Epsilon * darts_derivative
                for darts_derivative in darts_derivatives
            ]
            fin_diff_part = self._create_darts_derivatives(
                var_list=inner_grad.state, darts_derivatives=darts_derivaives
            )
            self._diff_initializer = tf.group(
                self._diff_initializer,
                tf.variables_initializer(fin_diff_part),
                tf.variables_initializer(grads_outer),
            )

            right_diff_0 = dict(
                zip(
                    model.task_parameter.keys(),
                    [
                        tf.add(state, fin_diff)
                        for state, fin_diff in zip(
                            model.task_parameter.values(), fin_diff_part
                        )
                    ],
                )
            )
            left_diff_0 = dict(
                zip(
                    model.task_parameter.keys(),
                    [
                        tf.subtract(state, fin_diff)
                        for state, fin_diff in zip(
                            model.task_parameter.values(), fin_diff_part
                        )
                    ],
                )
            )

            left_diff = tf.gradients(
                loss_func(
                    pred=model.re_forward(task_parameter=left_diff_0).out,
                    label=ex.y,
                    method="MetaRepr",
                ),
                xs=meta_param,
            )
            right_diff = tf.gradients(
                loss_func(
                    pred=model.re_forward(task_parameter=right_diff_0).out,
                    label=ex.y,
                    method="MetaRepr",
                ),
                xs=meta_param,
            )

            # compute the second-order part and add them to the first-order item
            for grad_outer, left_dif, right_dif in zip(
                grads_outer, left_diff, right_diff
            ):
                if right_dif is not None and left_dif is not None:
                    grad_param = tf.divide(
                        tf.subtract(right_dif, left_dif), 2 * self.epsilon
                    )
                    meta_grad = self.param_dict["learning_rate"] * grad_param
                    self._darts_initializer = tf.group(
                        self._darts_initializer, grad_outer.assign_sub(meta_grad)
                    )

            for h, doo_dh in zip(meta_param, grads_outer):
                assert doo_dh is not None, BOMLOuterGrad._ERROR_HYPER_DETACHED.format(
                    doo_dh
                )
                self._hypergrad_dictionary[h].append(doo_dh)
            return meta_param

    @staticmethod
    def _create_darts_derivatives(var_list, darts_derivatives):
        derivatives = [
            slot_creator.create_slot(
                v.initialized_value(), utils.val_or_zero(der, v), "alpha"
            )
            for v, der in zip(var_list, darts_derivatives)
        ]
        [
            tf.add_to_collection(boml.extension.GraphKeys.DARTS_DERIVATIVES, der)
            for der in derivatives
        ]
        boml.extension.remove_from_collection(
            boml.extension.GraphKeys.GLOBAL_VARIABLES, *derivatives
        )
        # this prevents the 'automatic' initialization with tf.global_variables_initializer.
        return derivatives

    @staticmethod
    def _create_outergradient_from_dodh(hyper, doo_dhypers):
        """
        Creates one hyper-gradient as a variable. doo_dhypers:  initialization, that is the derivative of
        the outer objective w.r.t this hyper
        """
        hgs = slot_creator.create_slot(
            hyper, utils.val_or_zero(doo_dhypers, hyper), "outergradient"
        )
        boml.extension.remove_from_collection(
            boml.extension.GraphKeys.GLOBAL_VARIABLES, hgs
        )
        return hgs

    @staticmethod
    def _create_outergradient(outer_obj, hyper):
        return BOMLOuterGradDarts._create_outergradient_from_dodh(
            hyper, tf.gradients(outer_obj, hyper)[0]
        )

    def apply_gradients(
        self,
        inner_objective_feed_dicts=None,
        outer_objective_feed_dicts=None,
        initializer_feed_dict=None,
        param_dict=OrderedDict(),
        global_step=None,
        session=None,
    ):
        if self._inner_method == "Aggr":
            alpha = param_dict["alpha"]
            t_tensor = param_dict["t_tensor"]

        ss = session or tf.get_default_session()

        self._history.clear()

        _fd = utils.maybe_call(initializer_feed_dict, utils.maybe_eval(global_step, ss))
        self._save_history(ss.run(self.initialization, feed_dict=_fd))

        # perform one-step update to the task parameters and store  weights along the optimization path
        _fd = inner_objective_feed_dicts
        if self._inner_method == "Aggr":
            _fd.update(outer_objective_feed_dicts)
            if not alpha.get_shape().as_list():
                _fd[t_tensor] = float(1.0)
            else:
                tmp = np.zeros((alpha.get_shape().as_list()[1], 1))
                tmp[0][0] = 1.0
                _fd[t_tensor] = tmp
        self._save_history(ss.run(self.iteration, feed_dict=_fd))

        # compute the differentiation part, multiplied by Epsilon with one-step forward pass
        _fd = utils.maybe_call(
            outer_objective_feed_dicts, utils.maybe_eval(global_step, ss)
        )
        darts_init_fd = utils.merge_dicts(_fd, inner_objective_feed_dicts)
        ss.run(self._diff_initializer, feed_dict=darts_init_fd)

        del self._history[-1]  # do not consider the final task parameters

        # compute the second-order part and add them to the first-order item
        state_feed_dict = utils.merge_dicts(
            *[
                od.state_feed_dict(h)
                for od, h in zip(sorted(self._optimizer_dicts), self._history[-1])
            ]
        )
        new_fd = utils.merge_dicts(state_feed_dict, inner_objective_feed_dicts)
        if self._inner_method == "Aggr":
            new_fd = utils.merge_dicts(new_fd, outer_objective_feed_dicts)
            # modified - mark
            if not alpha.shape.as_list():
                new_fd[t_tensor] = float(1.0)
            else:
                tmp = np.zeros((alpha.get_shape().as_list()[1], 1))
                tmp[0][0] = 1
                new_fd[t_tensor] = tmp
        new_fd = utils.merge_dicts(new_fd, outer_objective_feed_dicts)
        ss.run(self._darts_initializer, new_fd)

    def _save_history(self, weights):
        self._history.append(weights)



"""
Contains some misc utility functions
"""

from functools import reduce

import numpy as np
import tensorflow as tf
from tensorflow.contrib import layers


def conv_block(boml_net, cweight, bweight=None):

    """ Perform, conv, batch norm, nonlinearity, and max pool
    :param boml_net: instance of BOMLNet
    :param cweight: parameter of convolutional filter
    :param bweight: bias for covolutional filter
    """
    if boml_net.max_pool:
        conv_out = tf.add(
            tf.nn.conv2d(boml_net.out, cweight, boml_net.no_stride, "SAME"), bweight
        )
    else:
        conv_out = tf.add(
            tf.nn.conv2d(boml_net.out, cweight, boml_net.stride, "SAME"), bweight
        )

    if boml_net.batch_norm is not None:
        batch_out = boml_net.batch_norm(
            inputs=conv_out,
            activation_fn=boml_net.activation,
            variables_collections=boml_net.var_collections,
        )
    else:
        batch_out = boml_net.activation(conv_out)
    if boml_net.max_pool:
        final_out = tf.nn.max_pool(batch_out, boml_net.stride, boml_net.stride, "VALID")
        return final_out
    else:
        return batch_out


def conv_block_t(boml_net, conv_weight, conv_bias, zweight):
    """ Perform, conv, batch norm, nonlinearity, and max pool
    :param boml_net: instance of BOMLNet
    :param convweight: parameter of convolutional filter
    :param conv_bias: bias for covolutional filter
    :param zweight: parameters of covolutional filter for t-layer"""
    if boml_net.max_pool:
        conv_out = tf.add(
            tf.nn.conv2d(boml_net.out, conv_weight, boml_net.no_stride, "SAME"),
            conv_bias,
        )
    else:
        conv_out = tf.add(
            tf.nn.conv2d(boml_net.out, conv_weight, boml_net.stride, "SAME"), conv_bias
        )
    conv_output = tf.nn.conv2d(conv_out, zweight, boml_net.no_stride, "SAME")

    if boml_net.batch_norm is not None:
        batch_out = layers.batch_norm(
            conv_output,
            activation_fn=boml_net.activation,
            variables_collections=boml_net.var_collections,
        )
    else:
        batch_out = boml_net.activation(conv_output)
    if boml_net.max_pool:
        final_out = tf.nn.max_pool(batch_out, boml_net.stride, boml_net.stride, "VALID")
        return final_out
    else:
        return batch_out


def conv_block_warp(boml_net, cweight, bweight, zweight, zbias):
    """ Perform, conv, batch norm, nonlinearity, and max pool
    :param boml_net: instance of BOMLNet
    :param convweight: parameter of convolutional filter
    :param conv_bias: bias for covolutional filter
    :param zweight: parameters of covolutional filter for warp-layer
    :param zbias: bias of covolutional filter for warp-layer"""
    if boml_net.max_pool:
        conv_out = tf.add(
            tf.nn.conv2d(boml_net.out, cweight, boml_net.no_stride, "SAME"), bweight
        )
    else:
        conv_out = tf.add(
            tf.nn.conv2d(boml_net.out, cweight, boml_net.stride, "SAME"), bweight
        )

    conv_output = tf.add(
        tf.nn.conv2d(conv_out, zweight, boml_net.no_stride, "SAME"), zbias
    )

    if boml_net.batch_norm is not None:
        batch_out = layers.batch_norm(
            conv_output,
            activation_fn=boml_net.activation,
            variables_collections=boml_net.var_collections,
        )
    else:
        batch_out = boml_net.activation(conv_output)
    if boml_net.max_pool:
        final_out = tf.nn.max_pool(batch_out, boml_net.stride, boml_net.stride, "VALID")
        return final_out
    else:
        return batch_out


def get_conv_weight(boml_net, i, initializer):
    """
    :param boml_net: instance of BOMLNet
    :param i: int32, the i-th layer to be defined
    :param initializer: function to initialize the weights of convolutional filter
    :return: created convolutional weights
    """
    if i == 0:
        return tf.get_variable(
            "conv" + str(i),
            [
                boml_net.kernel,
                boml_net.kernel,
                boml_net.channels,
                boml_net.dim_hidden[0],
            ],
            initializer=initializer,
            dtype=boml_net.datatype,
        )
    else:
        return tf.get_variable(
            "conv" + str(i),
            [
                boml_net.kernel,
                boml_net.kernel,
                boml_net.dim_hidden[i - 1],
                boml_net.dim_hidden[i],
            ],
            initializer=initializer,
            dtype=boml_net.datatype,
        )


def get_bias_weight(boml_net, i, initializer):
    """
    :param boml_net: instance of BOMLNet
    :param i: int32, i-th layer to be defined
    :param initializer: function to initialize the bias
    :return: created convolutional bias
    """
    return tf.get_variable(
        "bias" + str(i),
        [boml_net.dim_hidden[i]],
        initializer=initializer,
        dtype=boml_net.datatype,
    )


def get_warp_weight(boml_net, layer, initializer):
    """
    :param boml_net: instance of BOMLNet
    :param layer: int32, i-th layer to be defined
    :param initializer: function to initializer the weights of convolutional filter
    :return: created convolutional parameters of Warp-Layer
    """
    return tf.get_variable(
        "conv" + str(layer) + "_z",
        [
            boml_net.kernel,
            boml_net.kernel,
            boml_net.dim_hidden[layer - 1],
            boml_net.dim_hidden[layer],
        ],
        initializer=initializer,
        dtype=boml_net.datatype,
    )


def get_warp_bias(boml_net, layer, initializer):
    """
    :param boml_net: instance of BOMLNet
    :param layer: int32, i-th layer to be defined
    :param initializer: function to initializer the weights of convolutional filter
    :return: created bias of convolutional filter
    """
    return tf.get_variable(
        "bias" + str(layer) + "_z",
        [boml_net.dim_hidden[layer]],
        initializer=initializer,
        dtype=boml_net.datatype,
    )


def get_identity(dim, name, conv=True):
    """
    :param dim: dimension of the identity matrix
    :param name: name for the variable
    :param conv: BOOLEAN, define variables as matrix or vector
    :return: created parameters of identity for T-layer
    """
    return (
        tf.Variable(tf.eye(dim, batch_shape=[1, 1]), name=name)
        if conv
        else tf.Variable(tf.eye(dim), name=name)
    )


def as_list(obj):
    """
    Makes sure `obj` is a list or otherwise converts it to a list with a single element.

    :param obj:
    :return: A `list`
    """
    return obj if isinstance(obj, list) else [obj]


def maybe_call(obj, *args, **kwargs):
    """
    Calls obj with args and kwargs and return its result if obj is callable, otherwise returns obj.
    """
    if callable(obj):
        return obj(*args, **kwargs)
    return obj


def as_tuple_or_list(obj):
    """
    Make sure that `obj` is a tuple or a list and eventually converts it into a list with a single element

    :param obj:
    :return: A `tuple` or a `list`
    """
    return obj if isinstance(obj, (list, tuple)) else [obj]


def maybe_get(obj, i):
    """

    :param obj: object
    :param i: the index
    :return: the i-th item if the `obj` instantiates the __getitem__ function
    """
    return obj[i] if hasattr(obj, "__getitem__") else obj


def merge_dicts(*dicts):
    return reduce(lambda a, nd: {**a, **nd}, dicts, {})


def to_one_hot_enc(seq, dimension=None):
    da_max = dimension or int(np.max(seq)) + 1
    _tmp = np.zeros((len(seq), da_max))
    _tmp[range(len(_tmp)), np.array(seq, dtype=int)] = 1
    return _tmp


def flatten_list(lst):
    from itertools import chain

    return list(chain(*lst))


def filter_vars(var_name, scope):
    import tensorflow as tf

    return [
        v
        for v in tf.get_collection(
            tf.GraphKeys.GLOBAL_VARIABLES,
            scope=scope.name if hasattr(scope, "name") else scope,
        )
        if v.name.endswith("%s:0" % var_name)
    ]


def name_from_vars(var_dict, *vars_):
    """
    Unfortunately this method doesn't return a very specific name....It gets a little messy

    :param var_dict:
    :param vars_:
    :return:
    """
    new_k_v = {}
    for v in vars_:
        for k, vv in var_dict.items():
            if v == vv:
                new_k_v[k] = v
    return name_from_dict(new_k_v)


def name_from_dict(_dict, *exclude_names):
    string_dict = {str(k): str(v) for k, v in _dict.items() if k not in exclude_names}
    return _tf_string_replace("_".join(flatten_list(list(sorted(string_dict.items())))))


def _tf_string_replace(_str):
    """
    Replace chars that are not accepted by tensorflow namings (eg. variable_scope)

    :param _str:
    :return:
    """
    return (
        _str.replace("[", "p")
        .replace("]", "q")
        .replace(",", "c")
        .replace("(", "p")
        .replace(")", "q")
        .replace(" ", "")
    )


def get_rand_state(rand):
    """
    Utility methods for getting a `RandomState` object.

    :param rand: rand can be None (new State will be generated),
                    np.random.RandomState (it will be returned) or an integer (will be treated as seed).

    :return: a `RandomState` object
    """
    if isinstance(rand, np.random.RandomState):
        return rand
    elif isinstance(rand, (int, np.ndarray, list)) or rand is None:
        return np.random.RandomState(rand)
    else:
        raise ValueError("parameter rand {} has wrong type".format(rand))


# SOME SCORING UTILS FUNCTIONS

half_int = lambda _m: 1.96 * np.std(_m) / np.sqrt(len(_m) - 1)


def mean_std_ci(measures, mul=1.0, tex=False):
    """
    Computes mean, standard deviation and 95% half-confidence interval for a list of measures.

    :param measures: list
    :param mul: optional multiplication coefficient (e.g. for percentage)
    :param tex: if True returns mean +- half_conf_interval for latex
    :return: a list or a string in latex
    """
    measures = np.array(measures) * mul
    ms = np.mean(measures), np.std(measures), half_int(measures)
    return ms if not tex else r"${:.2f} \pm {:.2f}$".format(ms[0], ms[2])


def leaky_relu(x, alpha, name=None):
    """
    Implements leaky relu with negative coefficient `alpha`
    """
    import tensorflow as tf

    with tf.name_scope(name, "leaky_relu_{}".format(alpha)):
        return tf.nn.relu(x) - alpha * tf.nn.relu(-x)






from boml.setup_model import network
from boml.setup_model.feedforward import BOMLNetFeedForward
from boml.setup_model.meta_init_v1 import (
    BOMLNetMetaInitV1,
    BOMLNetMiniMetaInitV1,
    BOMLNetOmniglotMetaInitV1,
)
from boml.setup_model.meta_init_v2 import (
    BOMLNetMiniMetaInitV2,
    BOMLNetOmniglotMetaInitV2,
)
from boml.setup_model.meta_repr_v2 import (
    BOMLNetMiniMetaReprV2,
    BOMLNetOmniglotMetaReprV2,
)
from boml.setup_model.meta_repr_v1 import (
    BOMLNetMetaReprV1,
    BOMLNetMiniMetaReprV1,
    BOMLNetOmniglotMetaReprV1,
)



# from progressbar import ProgressBar, Percentage, Bar, ETA
import os
import sys

import h5py
import numpy as np
from PIL import Image

from boml.load_data.datasets.load_full_dataset import MINI_IMAGENET_FOLDER_RES84


def img_to_array_old(image_file, resize=84, flip=False, convert_to_gray=False):
    img_z = Image.open(image_file)
    # img_z.convert('RGB')  # not useful at all...
    if resize:
        img_z = img_z.resize((resize, resize), Image.LANCZOS)
    if (
        flip
    ):  # when reading a stack saved in vaa3d format Y coordinates are reversed (used by save_subst
        img_z = img_z.transpose(Image.FLIP_TOP_BOTTOM)
    if convert_to_gray:
        img_z = img_z.convert("L")
    ary = np.array(img_z)
    ary = (
        np.stack((ary,) * 3, axis=2) if ary.ndim == 2 else ary
    )  # some images are black and white
    return (
        ary if ary.shape[-1] == 3 else ary[:, :, :3]
    )  # some images have an alpha channel... (unbelievable...)


def img_to_array(imfile, resize=84):
    from scipy.misc import imresize, imread

    img = imread(imfile, mode="RGB")
    if resize:
        # noinspection PyTypeChecker
        img = imresize(img, size=(resize, resize, 3))
    return img


def images_to_tensor(files, tensor):
    # widgets = [Percentage(), ' ', Bar(), ' ', ETA()]
    # pbar = ProgressBar(widgets=widgets, maxval=len(files))
    # pbar.start()
    for i, f in enumerate(files):
        print(f)
        tensor[i, :] = img_to_array(f)
        # pbar.update(i + 1)
        # pbar.finish()


def tensor_to_images(prefix, tensor):
    n = tensor.shape[0]
    # widgets = [Percentage(), ' ', Bar(), ' ', ETA()]
    # pbar = ProgressBar(widgets=widgets, maxval=n)
    # pbar.start()
    for z in range(n):
        img = Image.fromarray(tensor[z].astype(np.int8), mode="L")
        img.save(prefix + "%03d" % z + ".tif")
        # pbar.update(z + 1)
        # pbar.finish()


def convert_mini_imagenet(_folder=None, resize=None):
    for st in ["train", "val", "test"]:
        folder = os.path.join(_folder or MINI_IMAGENET_FOLDER_RES84, st)
        classes = os.listdir(folder)
        files = []
        for c in classes:
            files += [
                os.path.join(folder, os.path.join(c, f))
                for f in os.listdir(os.path.join(folder, c))
            ]
        # print(files)
        n = len(files)
        img_shape = img_to_array(files[0], resize=resize).shape

        img = img_to_array(files[0], resize=resize)
        print(img)
        print(img.dtype)

        h5file = h5py.File(
            os.path.join(_folder or MINI_IMAGENET_FOLDER_RES84, "%s.h5" % st), "w"
        )
        X = h5file.create_dataset(
            "X", (n, img_shape[0], img_shape[1], img_shape[2]), dtype=np.uint8
        )
        images_to_tensor(files, X)


if __name__ == "__main__":
    convert_mini_imagenet(sys.argv[1] if len(sys.argv) > 1 else None, resize=None)
    # base = '../vasculature_data'
    # orig_files = ['{}/orig/097000_164000_08600000{:0>2d}.tif'.format(base, z) for z in range(100)]
    # target_files = ['{}/GT/097000_164000_086000-labels-stack-intero-smooth00{:0>2d}.tif'.format(base, z) for z in range(100)]
    #
    # assert(len(orig_files) == len(target_files))
    # n = len(orig_files)
    # img_shape = img_to_array(orig_files[0]).shape
    #
    # h5file = h5py.File('vasculature.h5', 'w')
    # X = h5file.create_dataset("X", (n,img_shape[0],img_shape[1]), h5py.h5t.NATIVE_FLOAT)
    # y = h5file.create_dataset("y", (n,img_shape[0],img_shape[1]), h5py.h5t.NATIVE_FLOAT)
    # images_to_tensor(orig_files, X, 65535.)
    # images_to_tensor(target_files, y, 255.)
    # h5file.close()
    #
    # base = '../vasculature_data'
    # orig_files = ['{}/orig/097000_164000_08600000{:0>2d}.tif'.format(base, z) for z in range(100)]
    # target_files = ['{}/GT/097000_164000_086000-labels-stack-intero-smooth00{:0>2d}.tif'.format(base, z) for z in
    #                 range(100)]



import sys

import numpy as np
import tensorflow as tf

from boml import utils

Meta_Repr_Method = ["MAML", "FOMAML", "Meta-SGD", "MT-net"]
Meta_Init_Method = ["Reverse", "Truncated", "DARTS", "BDA", "Implicit"]
METHOD_COLLECTIONS = [Meta_Repr_Method, Meta_Init_Method]


class GraphKeys(tf.GraphKeys):
    """
    adds some meta_parameters and outer_gradients computation related keys
    """

    METAPARAMETERS = "meta_parameters"
    MODELPARAMETERS = "model_parameters"
    LAGRANGIAN_MULTIPLIERS = "lagrangian_multipliers"
    OUTERGRADIENTS = "outergradients"
    DARTS_DERIVATIVES = "darts_derivatives"
    ZS = "zs"


METAPARAMETERS_COLLECTIONS = [GraphKeys.METAPARAMETERS, GraphKeys.GLOBAL_VARIABLES]


def lagrangian_multipliers(scope=None):
    """
    List of variables in the collection LAGRANGIAN_MULTIPLIERS.

    These variables are created by `far.ReverseHG`.

    :param scope: (str) an optional scope.
    :return: A list of tensors (usually variables)
    """
    return tf.get_collection(GraphKeys.LAGRANGIAN_MULTIPLIERS, scope=scope)


def hypergradients(scope=None):
    """
    List of tensors and/or variables in the collection OUTERGRADIENTS.

    These variables are created by `far.HyperGradient`.

    :param scope: (str) an optional scope.
    :return: A list of tensors (usually variables)
    """
    return tf.get_collection(GraphKeys.OUTERGRADIENTS, scope=scope)


def remove_from_collection(key, *lst):
    """
    Remove tensors in lst from collection given by key
    """
    try:
        # noinspection PyProtectedMember
        [tf.get_default_graph()._collections[key].remove(_e) for _e in lst]
    except ValueError:
        print(
            "WARNING: Collection -> {} <- does not contain some tensor in {}".format(
                key, lst
            ),
            file=sys.stderr,
        )


def outer_parameters(scope=None):
    """
    List of variables in the collection HYPERPARAMETERS.

    Hyperparameters constructed with `get_outerparameter` are in this collection by default.

    :param scope: (str) an optional scope.
    :return: A list of tensors (usually variables)
    """
    return tf.get_collection(GraphKeys.METAPARAMETERS, scope=scope)


def get_outerparameter(
    name,
    initializer=None,
    shape=None,
    dtype=None,
    collections=None,
    scalar=False,
    constraint=None,
):
    """
    Creates an hyperparameter variable, which is a GLOBAL_VARIABLE
    and HYPERPARAMETER. Mirrors the behavior of `tf.get_variable`.

    :param name: name of this hyperparameter
    :param initializer: initializer or initial value (can be also np.array or float)
    :param shape: optional shape, may be not needed depending on initializer
    :param dtype: optional type,  may be not needed depending on initializer
    :param collections: optional additional collection or list of collections, which will be added to
                        METAPARAMETER and GLOBAL_VARIABLES
    :param scalar: default False, if True splits the hyperparameter in its scalar components, i.e. each component
                    will be a single scalar hyperparameter. In this case the method returns a tensor which of the
                    desired shape (use this option with `ForwardHG`)
    :param constraint: optional contstraint for the variable (only if not scalar..)

    :return: the newly created variable, or, if `scalar` is `True` a tensor composed by scalar variables.
    """
    _coll = list(METAPARAMETERS_COLLECTIONS)
    if collections:
        _coll += utils.as_list(collections)
    if not scalar:
        try:
            return tf.get_variable(
                name,
                shape,
                dtype,
                initializer,
                trainable=False,
                collections=_coll,
                constraint=constraint,
            )
        except TypeError as e:
            print(e)
            print("Trying to ignore constraints (to use constraints update tensorflow.")
            return tf.get_variable(
                name, shape, dtype, initializer, trainable=False, collections=_coll
            )
    else:
        with tf.variable_scope(name + "_components"):
            _shape = shape or initializer.shape
            if isinstance(_shape, tf.TensorShape):
                _shape = _shape.as_list()
            _tmp_lst = np.empty(_shape, object)
            for k in range(np.multiply.reduce(_shape)):
                indices = np.unravel_index(k, _shape)
                _ind_name = "_".join([str(ind) for ind in indices])
                _tmp_lst[indices] = tf.get_variable(
                    _ind_name,
                    (),
                    dtype,
                    initializer if callable(initializer) else initializer[indices],
                    trainable=False,
                    collections=_coll,
                )
        return tf.convert_to_tensor(_tmp_lst.tolist(), name=name)


def get_global_step(name="GlobalStep", init=0):
    import tensorflow as tf

    return tf.get_variable(
        name,
        initializer=init,
        trainable=False,
        collections=[tf.GraphKeys.GLOBAL_STEP, tf.GraphKeys.GLOBAL_VARIABLES],
    )


import os
from functools import reduce

import numpy as np
import scipy as sp
import scipy.sparse as sc_sp
import tensorflow as tf


def get_indices_balanced_classes(n_examples, labels, forbidden_indices):
    N = len(labels)
    n_classes = len(labels[0])

    indices = []
    current_class = 0
    for i in range(n_examples):
        index = np.random.random_integers(0, N - 1, 1)[0]
        while (
            index in indices
            or index in forbidden_indices
            or np.argmax(labels[index]) != current_class
        ):
            index = np.random.random_integers(0, N - 1, 1)[0]
        indices.append(index)
        current_class = (current_class + 1) % n_classes

    return indices


def test_if_balanced(dataset):
    labels = dataset.target
    n_classes = len(labels[0])
    class_counter = [0] * n_classes
    for l in labels:
        class_counter[np.argmax(l)] += 1
    print("exemple by class: ", class_counter)


def maybe_cast_to_scalar(what):
    return what[0] if len(what) == 1 else what


def pad(_example, _size):
    return np.concatenate([_example] * _size)


def stack_or_concat(list_of_arays):
    func = np.concatenate if list_of_arays[0].ndim == 1 else np.vstack
    return func(list_of_arays)


def merge_dicts(*dicts):
    return reduce(lambda a, nd: {**a, **nd}, dicts, {})


def vstack(lst):
    """
    Vstack that considers sparse matrices

    :param lst:
    :return:
    """
    return (
        sp.vstack(lst)
        if sp and isinstance(lst[0], sc_sp.csr.csr_matrix)
        else np.vstack(lst)
    )


def convert_sparse_matrix_to_sparse_tensor(X):
    if isinstance(X, sc_sp.csr.csr_matrix):
        coo = X.tocoo()
        indices = np.mat([coo.row, coo.col]).transpose()
    else:
        coo, indices = X, [X.row, X.col]
    # data = np.array(coo.data, dtype=)
    return tf.SparseTensor(indices, tf.constant(coo.data, dtype=tf.float32), coo.shape)


def get_data(d_set):
    if hasattr(d_set, "images"):
        data = d_set.images
    elif hasattr(d_set, "data"):
        data = d_set.data
    else:
        raise ValueError("something wrong with the dataset %s" % d_set)
    return data


def get_targets(d_set):
    if hasattr(d_set, "labels"):
        return d_set.labels
    elif hasattr(d_set, "target"):
        return d_set.target
    else:
        raise ValueError("something wrong with the dataset %s" % d_set)


def as_list(obj):
    """
    Makes sure `obj` is a list or otherwise converts it to a list with a single element.

    :param obj:
    :return: A `list`
    """
    return obj if isinstance(obj, list) else [obj]


def merge_dicts(*dicts):
    return reduce(lambda a, nd: {**a, **nd}, dicts, {})


def get_rand_state(rand):
    """
    Utility methods for getting a `RandomState` object.

    :param rand: rand can be None (new State will be generated),
                    np.random.RandomState (it will be returned) or an integer (will be treated as seed).

    :return: a `RandomState` object
    """
    if isinstance(rand, np.random.RandomState):
        return rand
    elif isinstance(rand, (int, np.ndarray, list)) or rand is None:
        return np.random.RandomState(rand)
    else:
        raise ValueError("parameter rand {} has wrong type".format(rand))


def maybe_call(obj, *args, **kwargs):
    """
    Calls obj with args and kwargs and return its result if obj is callable, otherwise returns obj.
    """
    if callable(obj):
        return obj(*args, **kwargs)
    return obj


def as_tuple_or_list(obj):
    """
    Make sure that `obj` is a tuple or a list and eventually converts it into a list with a single element

    :param obj:
    :return: A `tuple` or a `list`
    """
    return obj if isinstance(obj, (list, tuple)) else [obj]


def to_one_hot_enc(seq, dimension=None):
    da_max = dimension or int(np.max(seq)) + 1
    _tmp = np.zeros((len(seq), da_max))
    _tmp[range(len(_tmp)), np.array(seq, dtype=int)] = 1
    return _tmp
    #
    # def create_and_set(_p):
    #     _tmp = np.zeros(da_max)
    #     _tmp[int(_p)] = 1
    #     return _tmp
    #
    # return np.array([create_and_set(_v) for _v in seq])


from_env = os.getenv("DATASETS_FOLDER")
DATA_FOLDER = from_env



from collections import OrderedDict

import tensorflow as tf

from boml.lower_iter import BOMLInnerGradTrad
from boml.utils import add_list


class BOMLInnerGradSimple(BOMLInnerGradTrad):
    def __init__(
        self,
        update_op,
        dynamics,
        objective,
        inner_param_tensor=None,
        outer_param_tensor=[],
        model_param_tensor=[],
    ):
        self._inner_param_fast_tensor = inner_param_tensor
        self._outer_param_tensor = outer_param_tensor
        self._model_param_tensor = model_param_tensor
        super().__init__(update_op=update_op, dynamics=dynamics, objective=objective)

    @staticmethod
    def compute_gradients(
        boml_opt,
        loss_inner,
        loss_outer=None,
        param_dict=OrderedDict(),
        var_list=None,
        **inner_kargs
    ):
        """
        :param boml_opt: instance of modified optimizers in the `optimizer` module
        :param loss_inner: Lower-Level objectives
        :param loss_outer: Upper-Level objectives
        :param param_dict: dictionary of general parameters for different algorithms
        :param var_list: the list of parameters in the base-learner
        :param inner_kargs: optional arguments for tensorflow optimizers, like global_step, gate_gradients
        :return: initialized instance of inner_grad for UL optimization
        """
        minimize_kargs = {
            inner_arg: inner_kargs[inner_arg]
            for inner_arg in set(inner_kargs.keys()) - set(param_dict.keys())
        }

        assert loss_inner is not None, "argument:inner_objective must be initialized"
        update_op, dynamics = boml_opt.minimize(loss_inner, var_list, *minimize_kargs)
        fast_param, outer_grad, model_grad = BOMLInnerGradSimple.bml_inner_grad_trad(
            loss_inner=loss_inner, param_dict=param_dict, var_list=var_list
        )

        return BOMLInnerGradSimple(
            update_op=update_op,
            dynamics=dynamics,
            objective=loss_inner,
            inner_param_tensor=fast_param,
            outer_param_tensor=outer_grad,
            model_param_tensor=model_grad,
        )

    @staticmethod
    def bml_inner_grad_trad(loss_inner, param_dict=OrderedDict(), var_list=[]):
        """

        :param loss_inner: LL objectives
        :param param_dict: dictionary of general parameters for different algorithms
        :param var_list: the list of parameters in the base-learner
        :return:
        """
        assert len(var_list) > 0, "no task_specific variables to optimizer"
        task_model = param_dict["experiment"].model
        task_param = task_model.task_parameter
        outer_param_grad = []
        model_param_grad = []
        grads = tf.gradients(loss_inner, list(task_param.values()))
        if param_dict["use_Warp"]:
            outer_param_loss = param_dict["outer_loss_func"](
                pred=task_model.re_forward(param_dict["experiment"].x_).out,
                label=param_dict["experiment"].y_,
                method="MetaInit",
            )
            model_param_loss = param_dict["model_loss_func"](
                pred=task_model.re_forward(param_dict["experiment"].x_).out,
                label=param_dict["experiment"].y_,
                method="MetaInit",
            )
            outer_param_grad = tf.gradients(outer_param_loss, list(var_list))
            model_param_grad = tf.gradients(
                model_param_loss, list(task_model.model_param_dict.values())
            )

        if param_dict["first_order"]:
            grads_dict = dict(
                zip(
                    task_model.task_parameter.keys(),
                    [tf.stop_gradient(grad) for grad in grads],
                )
            )
        else:
            grads_dict = dict(zip(task_model.task_parameter.keys(), grads))

        task_model.task_parameter = OrderedDict(
            zip(
                task_param.keys(),
                [
                    task_param[key] - param_dict["learning_rate"] * grads_dict[key]
                    for key in task_param.keys()
                ],
            )
        )

        for _ in range(param_dict["T"] - 1):
            task_model = task_model.re_forward(param_dict["experiment"].x)
            task_param = task_model.task_parameter

            iter_loss = param_dict["loss_func"](
                pred=task_model.out, label=param_dict["experiment"].y, method="MetaInit"
            )

            if param_dict["use_Warp"]:
                outer_param_loss = param_dict["outer_loss_func"](
                    pred=task_model.re_forward(param_dict["experiment"].x_).out,
                    label=param_dict["experiment"].y_,
                    method="MetaInit",
                )
                model_param_loss = param_dict["model_loss_func"](
                    pred=task_model.re_forward(param_dict["experiment"].x_).out,
                    label=param_dict["experiment"].y_,
                    method="MetaInit",
                )
                outer_param_grad = add_list(
                    outer_param_grad, tf.gradients(outer_param_loss, list(var_list))
                )
                model_param_grad = add_list(
                    model_param_grad,
                    tf.gradients(
                        model_param_loss, list(task_model.model_param_dict.values())
                    ),
                )

            grads = tf.gradients(iter_loss, list(task_param.values()))

            if param_dict["first_order"]:
                grads_dict = dict(
                    zip(
                        task_model.task_parameter.keys(),
                        [tf.stop_gradient(grad) for grad in grads],
                    )
                )
            else:
                grads_dict = dict(zip(task_model.task_parameter.keys(), grads))

            task_model.task_parameter = OrderedDict(
                zip(
                    task_param.keys(),
                    [
                        task_param[key] - param_dict["learning_rate"] * grads_dict[key]
                        for key in task_param.keys()
                    ],
                )
            )
        param_dict["experiment"].model = task_model

        return task_model.task_parameter, outer_param_grad, model_param_grad

    @property
    def outer_param_tensor(self):
        """
        :return: temporary weights dictionary used in maml and fomaml for back propagation
        """
        assert (
            self._outer_param_tensor is not None
        ), "temporary weights dictionary must be initialized before being called"
        return self._outer_param_tensor

    @property
    def model_param_tensor(self):
        """
        :return: temporary weights dictionary used in maml and fomaml for back propagation
        """
        assert (
            self._model_param_tensor is not None
        ), "temporary weights dictionary must be initialized before being called"
        return self._model_param_tensor



import sys
from collections import OrderedDict

import tensorflow as tf
from tensorflow.python.training import slot_creator

from boml.extension import GraphKeys
from boml.setup_model.network_utils import filter_vars
from boml.utils import remove_from_collection


class BOMLNet(object):
    # definitely deprecated!
    """
    Base object for building neural networks
    """

    def __init__(
        self,
        _input,
        outer_param_dict=OrderedDict(),
        model_param_dict=OrderedDict(),
        var_collections=None,
        name=None,
        reuse=False,
    ):
        """
        Creates an object that creates model parameters and defines the network structure.
        :param _input: the input shape for defined network
        :param outer_param_dict: type of OrderedDict to restore parameters to optimize in the upper level
        :param model_param_dict: type of OrderedDict to restore parameters of special model like Iteration Warping
        :param task_parameter: type of OrderedDict to restore parameters to optimize in the lower level
        """
        super(BOMLNet, self).__init__()

        if not name:
            try:
                name = tf.get_variable_scope().name
            except IndexError:
                print("Warning: no name and no variable scope given", sys.stderr)
        self.outer_param_dict = outer_param_dict
        self.model_param_dict = model_param_dict
        self.var_collections = var_collections
        self.name = name
        self.reuse = reuse
        self._var_initializer_op = None

        self.layers = [_input]
        # self.s = None
        self._tf_saver = None

        with self._variable_scope(reuse):
            if (len(self.outer_param_dict) == 0) and (
                callable(getattr(self, "create_outer_parameters", None))
            ):
                self.create_outer_parameters()
            self._forward()

    def _variable_scope(self, reuse):
        """
        May override default variable scope context.
        :param reuse:
        :return:
        """
        return tf.variable_scope(self.name, reuse=reuse)

    def __getitem__(self, item):
        """
        Get's the `activation`

        :param item:
        :return:
        """
        return self.layers[item]

    def __add__(self, other):
        if self.name not in tf.get_variable_scope().name:
            print(
                "Warning: adding layers outside model variable scope", file=sys.stderr
            )
        self.layers.append(other)
        return self

    @property
    def var_list(self):
        """
        :return: list that contains the variables created in the current scope
        """
        return tf.get_collection(tf.GraphKeys.MODEL_VARIABLES, self.name)

    @property
    def out(self):
        """
        :return: the current output of the BOMLNet
        """
        return self[-1]

    def create_initial_parameter(self, primary_outerparameter=None):
        """
        :param primary_outerparameter: the primary outerparameters used to
        create the task-specific parameters
        :return: dictionary to keep the created task parameters
        """
        assert primary_outerparameter is not None, (
            "Primary hyperparameters "
            "must be provided for initialization of slot variables"
        )
        # task_weights_keys = [key for key in primary_hyperparameter.keys() if 'prob' not in key and 'z' not in key]
        initial_parameter = OrderedDict(
            [
                (
                    primary_key,
                    slot_creator.create_slot(
                        primary=self.out,
                        val=primary_outerparameter[primary_key].initialized_value(),
                        name=primary_key,
                    ),
                )
                for primary_key in primary_outerparameter.keys()
            ]
        )
        [
            tf.add_to_collection(self.var_collections, initial_param)
            for initial_param in initial_parameter.values()
        ]
        remove_from_collection(GraphKeys.GLOBAL_VARIABLES, *initial_parameter.values())
        return initial_parameter

    def filter_vars(self, var_name):
        return filter_vars(var_name, self.name)

    def initialize(self, session=None):
        """
        Initialize the model. If `deterministic_initialization` is set to true,
        saves the initial weight in numpy which will be used for subsequent initialization.
        This is because random seed management in tensorflow is rather obscure... and I could not
        find a way to set the same seed across different initialization without exiting the session.

        :param session:
        :return:
        """
        # TODO placeholder are not useful at all here... just change initializer of tf.Variables !
        ss = session or tf.get_default_session()
        assert ss, "No default session"
        if not self._var_initializer_op:
            self._var_initializer_op = tf.variables_initializer(self.var_list)
        ss.run(self._var_initializer_op)

    def _forward(self):
        """
        _forward() uses defined convolutional neural networks with initial input
        :return:
        """
        raise NotImplemented()

    def re_forward(self, new_input):
        """
        reuses defined convolutional networks with new input and update the output results
        :param new_input: new input with same shape as the old one
        :return: updated instance of BOMLNet
        """
        raise NotImplemented()



"""
Contains some misc utility functions
"""
import collections
import multiprocessing
from collections import OrderedDict, Callable
from functools import reduce

import numpy as np


def as_list(obj):
    """
    Makes sure `obj` is a list or otherwise converts it to a list with a single element.

    :param obj:
    :return: A `list`
    """
    return obj if isinstance(obj, list) else [obj]


def maybe_call(obj, *args, **kwargs):
    """
    Calls obj with args and kwargs and return its result if obj is callable, otherwise returns obj.
    """
    if callable(obj):
        return obj(*args, **kwargs)
    return obj


def as_tuple_or_list(obj):
    """
    Make sure that `obj` is a tuple or a list and eventually converts it into a list with a single element

    :param obj:
    :return: A `tuple` or a `list`
    """
    return obj if isinstance(obj, (list, tuple)) else [obj]


def maybe_get(obj, i):
    return obj[i] if hasattr(obj, "__getitem__") else obj


def merge_dicts(*dicts):
    return reduce(lambda a, nd: {**a, **nd}, dicts, {})


def flatten_list(lst):
    from itertools import chain

    return list(chain(*lst))


def filter_vars(var_name, scope):
    import tensorflow as tf

    return [
        v
        for v in tf.get_collection(
            tf.GraphKeys.GLOBAL_VARIABLES,
            scope=scope.name if hasattr(scope, "name") else scope,
        )
        if v.name.endswith("%s:0" % var_name)
    ]


def name_from_vars(var_dict, *vars_):
    """
    Unfortunately this method doesn't return a very specific name....It gets a little messy

    :param var_dict:
    :param vars_:
    :return:
    """
    new_k_v = {}
    for v in vars_:
        for k, vv in var_dict.items():
            if v == vv:
                new_k_v[k] = v
    return name_from_dict(new_k_v)


def name_from_dict(_dict, *exclude_names):
    string_dict = {str(k): str(v) for k, v in _dict.items() if k not in exclude_names}
    return _tf_string_replace("_".join(flatten_list(list(sorted(string_dict.items())))))


def _tf_string_replace(_str):
    """
    Replace chars that are not accepted by tensorflow namings (eg. variable_scope)

    :param _str:
    :return:
    """
    return (
        _str.replace("[", "p")
        .replace("]", "q")
        .replace(",", "c")
        .replace("(", "p")
        .replace(")", "q")
        .replace(" ", "")
    )


def namedtuple_with_defaults(typename, field_names, default_values=()):
    T = collections.namedtuple(typename, field_names)
    T.__new__.__defaults__ = (None,) * len(T._fields)
    if isinstance(default_values, collections.Mapping):
        prototype = T(**default_values)
    else:
        prototype = T(*default_values)
    T.__new__.__defaults__ = tuple(prototype)
    return T


def get_rand_state(rand):
    """
    Utility methods for getting a `RandomState` object.

    :param rand: rand can be None (new State will be generated),
                    np.random.RandomState (it will be returned) or an integer (will be treated as seed).

    :return: a `RandomState` object
    """
    if isinstance(rand, np.random.RandomState):
        return rand
    elif isinstance(rand, (int, np.ndarray, list)) or rand is None:
        return np.random.RandomState(rand)
    else:
        raise ValueError("parameter rand {} has wrong type".format(rand))


def GPU_CONFIG():
    import tensorflow as tf

    CONFIG_GPU_GROWTH = tf.ConfigProto(allow_soft_placement=True)
    CONFIG_GPU_GROWTH.gpu_options.allow_growth = True
    return CONFIG_GPU_GROWTH


# SOME SCORING UTILS FUNCTIONS

half_int = lambda _m: 1.96 * np.std(_m) / np.sqrt(len(_m) - 1)


def mean_std_ci(measures, mul=1.0, tex=False):
    """
    Computes mean, standard deviation and 95% half-confidence interval for a list of measures.

    :param measures: list
    :param mul: optional multiplication coefficient (e.g. for percentage)
    :param tex: if True returns mean +- half_conf_interval for latex
    :return: a list or a string in latex
    """
    measures = np.array(measures) * mul
    ms = np.mean(measures), np.std(measures), half_int(measures)
    return ms if not tex else r"${:.2f} \pm {:.2f}$".format(ms[0], ms[2])


def leaky_relu(x, alpha, name=None):
    """
    Implements leaky relu with negative coefficient `alpha`
    """
    import tensorflow as tf

    with tf.name_scope(name, "leaky_relu_{}".format(alpha)):
        return tf.nn.relu(x) - alpha * tf.nn.relu(-x)


def execute(target, *args, **kwargs):
    pr = multiprocessing.Process(target=target, args=args, kwargs=kwargs)
    pr.start()
    return pr


def get_global_step(name="GlobalStep", init=0):
    import tensorflow as tf

    return tf.get_variable(
        name,
        initializer=init,
        trainable=False,
        collections=[tf.GraphKeys.GLOBAL_STEP, tf.GraphKeys.GLOBAL_VARIABLES],
    )


class DefaultOrderedDict(OrderedDict):
    # Source: http://stackoverflow.com/a/6190500/562769
    def __init__(self, default_factory=None, *a, **kw):
        if default_factory is not None and not isinstance(default_factory, Callable):
            raise TypeError("first argument must be callable")
        OrderedDict.__init__(self, *a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = (self.default_factory,)
        return type(self), args, None, None, self.items()

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        import copy

        return type(self)(self.default_factory, copy.deepcopy(self.items()))

    def __repr__(self):
        return "OrderedDefaultDict(%s, %s)" % (
            self.default_factory,
            OrderedDict.__repr__(self),
        )



from __future__ import print_function, absolute_import, division

import tensorflow as tf


class BOMLOpt(tf.train.Optimizer):
    """
    mirror of the tf.train.Optimizer.
    """
    def minimize(
        self,
        loss_inner,
        var_list=None,
        global_step=None,
        gate_gradients=tf.train.Optimizer.GATE_OP,
        aggregation_method=None,
        colocate_gradients_with_ops=False,
        name=None,
        grad_loss=None,
    ):
        """ The `dynamics` contains a list of var_and_dynamics where var are both variables in `var_list` and also
        additional state (auxiliary) variables to be used in the process of back-propagation.
        """
        update_op, dynamics = super(BOMLOpt, self).minimize(
            loss_inner,
            global_step,
            var_list,
            gate_gradients,
            aggregation_method,
            colocate_gradients_with_ops,
            name,
            grad_loss,
        )
        return update_op, dynamics

    @property
    def learning_rate(self):
        return self._learning_rate

    @property
    def learning_rate_tensor(self):
        return self._learning_rate_tensor

    @property
    def optimizer_params_tensor(self):
        return [self.learning_rate_tensor]



from collections import OrderedDict
from functools import reduce

import numpy as np
import tensorflow as tf
from tensorflow.contrib.layers.python import layers

from boml import extension
from boml.setup_model import network_utils
from boml.setup_model.network import BOMLNet


class BOMLNetMetaReprV1(BOMLNet):
    def __init__(
        self,
        _input,
        name="BMLNetC4LMetaRepr",
        outer_param_dict=OrderedDict(),
        model_param_dict=OrderedDict(),
        task_parameter=None,
        use_T=False,
        use_Warp=False,
        outer_method="Reverse",
        dim_output=-1,
        activation=tf.nn.relu,
        var_collections=extension.METAPARAMETERS_COLLECTIONS,
        conv_initializer=tf.contrib.layers.xavier_initializer_conv2d(tf.float32),
        output_weight_initializer=tf.contrib.layers.xavier_initializer(tf.float32),
        norm=layers.batch_norm,
        data_type=tf.float32,
        channels=1,
        dim_hidden=[64, 64, 64, 64],
        kernel=3,
        max_pool=False,
        reuse=False,
    ):
        self.task_parameter = task_parameter
        self.dim_output = dim_output
        self.kernel = kernel
        self.channels = channels
        self.dim_hidden = dim_hidden
        self.datatype = data_type
        self.batch_norm = norm
        self.max_pool = max_pool
        self.stride = [1, 2, 2, 1]
        self.no_stride = [1, 1, 1, 1]
        self.activation = activation
        self.bias_initializer = tf.zeros_initializer(tf.float32)
        self.conv_initializer = conv_initializer
        self.output_weight_initializer = output_weight_initializer
        self.use_T = use_T
        self.use_Warp = use_Warp
        self.outer_method = outer_method
        self.flatten = False if self.outer_method == "Implicit" else True

        super(BOMLNetMetaReprV1, self).__init__(
            _input=_input,
            outer_param_dict=outer_param_dict,
            var_collections=var_collections,
            name=name,
            model_param_dict=model_param_dict,
            reuse=reuse,
        )

        self.betas = self.filter_vars("beta")

        self.moving_means = self.filter_vars("moving_mean")
        self.moving_variances = self.filter_vars("moving_variance")

        if not reuse:
            extension.remove_from_collection(
                extension.GraphKeys.MODEL_VARIABLES, *self.moving_means
            )
            extension.remove_from_collection(
                extension.GraphKeys.MODEL_VARIABLES, *self.moving_variances
            )
            print(name, "MODEL CREATED")
        extension.remove_from_collection(
            extension.GraphKeys.METAPARAMETERS, *self.moving_means
        )
        extension.remove_from_collection(
            extension.GraphKeys.METAPARAMETERS, *self.moving_variances
        )

    def create_outer_parameters(self):

        for i in range(len(self.dim_hidden)):
            self.outer_param_dict["conv" + str(i)] = network_utils.get_conv_weight(
                self, i=i, initializer=self.conv_initializer
            )

            self.outer_param_dict["bias" + str(i)] = network_utils.get_bias_weight(
                self, i=i, initializer=self.bias_initializer
            )
        [
            tf.add_to_collections(extension.GraphKeys.METAPARAMETERS, hyper)
            for hyper in self.outer_param_dict.values()
        ]

        if len(self.model_param_dict) == 0 and callable(
            getattr(self, "create_model_parameters", None)
        ):
            self.create_model_parameters()

        return self.outer_param_dict

    def create_model_parameters(
        self, var_collections=extension.GraphKeys.METAPARAMETERS
    ):
        if self.use_T:
            # hyper parameters of transformation layer
            for i in range(len(self.dim_hidden)):
                self.model_param_dict[
                    "conv" + str(i) + "_z"
                ] = network_utils.get_identity(
                    self.dim_hidden[0], name="conv" + str(i) + "_z", conv=True
                )
        elif self.use_Warp:
            for i in range(len(self.dim_hidden)):
                self.model_param_dict[
                    "conv" + str(i) + "_z"
                ] = network_utils.get_warp_weight(
                    self, i, self.conv_initializer
                )
                self.model_param_dict[
                    "bias" + str(i) + "_z"
                ] = network_utils.get_warp_bias(
                    self, i, self.bias_initializer
                )
        [
            tf.add_to_collections(var_collections, model_param)
            for model_param in self.model_param_dict.values()
        ]

        return self.model_param_dict

    def _forward(self):
        """
        for i in range(4):
            self.conv_layer(filters=self.dim_hidden[i],stride=self.stride, max_pool=self.max_pool)
        flattened_shape = reduce(lambda a, v: a * v, self.layers[-1].get_shape().as_list()[1:])
        self + tf.reshape(self.out, shape=(-1, flattened_shape), name='representation')
        """

        for i in range(len(self.dim_hidden)):
            if self.use_T:
                self + network_utils.conv_block_t(
                    self,
                    self.outer_param_dict["conv" + str(i)],
                    self.outer_param_dict["bias" + str(i)],
                    self.model_param_dict["conv" + str(i) + "_z"],
                )
            elif self.use_Warp:
                self + network_utils.conv_block_warp(
                    self,
                    self.outer_param_dict["conv" + str(i)],
                    self.outer_param_dict["bias" + str(i)],
                    self.model_param_dict["conv" + str(i) + "_z"],
                    self.model_param_dict["bias" + str(i) + "_z"],
                )
            else:
                self + network_utils.conv_block(
                    self,
                    self.outer_param_dict["conv" + str(i)],
                    self.outer_param_dict["bias" + str(i)],
                )
        if self.flatten:
            flattened_shape = reduce(
                lambda a, v: a * v, self.layers[-1].get_shape().as_list()[1:]
            )
            self + tf.reshape(
                self.out, shape=(-1, flattened_shape), name="representation"
            )
        else:
            if self.max_pool:
                self + tf.reshape(
                    self.out,
                    [-1, np.prod([int(dim) for dim in self.out.get_shape()[1:]])],
                )
            else:
                self + tf.reduce_mean(self.out, [1, 2])

    def re_forward(self, new_input):
        return BOMLNetMetaReprV1(
            _input=new_input if new_input is not None else self.layers[0],
            name=self.name,
            activation=self.activation,
            outer_param_dict=self.outer_param_dict,
            model_param_dict=self.model_param_dict,
            dim_output=self.dim_output,
            task_parameter=self.task_parameter,
            use_Warp=self.use_Warp,
            use_T=self.use_T,
            var_collections=self.var_collections,
            dim_hidden=self.dim_hidden,
            output_weight_initializer=self.output_weight_initializer,
            max_pool=self.max_pool,
            reuse=True,
            outer_method=self.outer_method,
        )


def BOMLNetOmniglotMetaReprV1(
    _input,
    outer_param_dict=OrderedDict(),
    model_param_dict=OrderedDict(),
    batch_norm=layers.batch_norm,
    name="BMLNetC4LOmniglot",
    use_T=False,
    dim_output=-1,
    use_Warp=False,
    outer_method="Reverse",
    **model_args
):

    return BOMLNetMetaReprV1(
        _input=_input,
        name=name,
        model_param_dict=model_param_dict,
        dim_output=dim_output,
        outer_param_dict=outer_param_dict,
        norm=batch_norm,
        use_T=use_T,
        use_Warp=use_Warp,
        outer_method=outer_method,
        **model_args
    )


def BOMLNetMiniMetaReprV1(
    _input,
    outer_param_dict=OrderedDict(),
    model_param_dict=OrderedDict(),
    dim_output=-1,
    batch_norm=layers.batch_norm,
    name="BOMLNetC4LMini",
    use_T=False,
    use_Warp=False,
    outer_method="Reverse",
    **model_args
):
    return BOMLNetMetaReprV1(
        _input=_input,
        name=name,
        use_T=use_T,
        use_Warp=use_Warp,
        dim_output=dim_output,
        outer_param_dict=outer_param_dict,
        model_param_dict=model_param_dict,
        norm=batch_norm,
        channels=3,
        dim_hidden=[32, 32, 32, 32],
        max_pool=True,
        outer_method=outer_method,
        **model_args
    )



from __future__ import print_function, absolute_import, division

from collections import OrderedDict

import tensorflow as tf

from boml.optimizer.opt import BOMLOpt

# import numpy as np

GRADIENT_NONE_MESSAGE = (
    "WARNING: the gradient w.r.t.the tf.Variable\n {}\n is None;\n "
    "Check the computational graph of the inner objective, and be sure you\n"
    "are not considering including variables that should not be there among the\n"
    "inner variables."
)


class BOMLOptSGD(BOMLOpt, tf.train.GradientDescentOptimizer):
    def __init__(self, learning_rate, use_locking=False, name="GradientDescent"):
        super(BOMLOptSGD, self).__init__(learning_rate, use_locking, name)

    def apply_gradients(self, grads_and_vars, global_step=None, name=None):
        update_op = super(BOMLOptSGD, self).apply_gradients(
            grads_and_vars, global_step, name
        )

        dynamics = OrderedDict()
        for g, w in grads_and_vars:
            assert g is not None, GRADIENT_NONE_MESSAGE.format(w)
            wk = w - tf.cast(self._learning_rate_tensor, g.dtype) * g
            dynamics[w] = wk
        return update_op, dynamics

    def __str__(self):
        return "{}-lr={}".format(self._name, self._learning_rate)



from __future__ import absolute_import, print_function, division

from collections import defaultdict, OrderedDict

import tensorflow as tf

import boml.extension
from boml.lower_iter import BOMLInnerGradTrad

RAISE_ERROR_ON_DETACHED = False


class BOMLOuterGrad(object):
    def __init__(self, name):
        self._optimizer_dicts = set()
        self._inner_objectives = None
        self._hypergrad_dictionary = defaultdict(list)
        self._apply_updates = None

        self._initialization = None
        self._iteration = None
        self._state = None
        self._name = name

    _ERROR_HYPER_DETACHED = """
    `The outer parameter` is detached from this optimization dynamics.
    """

    def compute_gradients(self, outer_objective, boml_inner_grad, meta_param=None):
        """
        Function overridden by specific methods.

        :param boml_inner_grad: inner_grad object resulting from the inner objective optimization.
        :param outer_objective: A loss function for the outer parameters
        :param meta_param: Optional list of outer parameters to consider. If not provided will get all variables in the
                            hyperparameter collection in the current scope.

        :return: list of outer parameters involved in the computation
        """
        assert isinstance(
            boml_inner_grad, BOMLInnerGradTrad
        ), BOMLOuterGrad._ERROR_NOT_OPTIMIZER_DICT.format(boml_inner_grad)
        self._optimizer_dicts.add(boml_inner_grad)

        if meta_param is None:  # get default outer parameters
            meta_param = boml.extension.meta_parameters(tf.get_variable_scope().name)
        return meta_param

    @property
    def initialization(self):
        if self._initialization is None:
            self._initialization = [
                opt_dict.initialization for opt_dict in sorted(self._optimizer_dicts)
            ]
        return self._initialization

    @property
    def iteration(self):
        if self._iteration is None:
            self._iteration = [
                opt_dict.iteration for opt_dict in sorted(self._optimizer_dicts)
            ]
        return self._iteration

    @property
    def state(self):
        for opt_dict in sorted(self._optimizer_dicts):
            for v in opt_dict.state:
                yield v

    @property
    def inner_objectives(self):
        if self._inner_objectives is None:
            self._inner_objectives = [
                opt.objective if hasattr(opt, "objective") else tf.constant(False)
                for opt in sorted(self._optimizer_dicts)
            ]
        return self._inner_objectives

    @property
    def apply_updates(self):
        if self._apply_updates is None:
            self._apply_updates = tf.group(
                *[opt_dict.apply_updates for opt_dict in sorted(self._optimizer_dicts)]
            )
        return self._apply_updates

    def apply_gradients(
        self,
        inner_objective_feed_dicts=None,
        outer_objective_feed_dicts=None,
        initializer_feed_dict=None,
        param_dict=OrderedDict(),
        global_step=None,
        session=None,
    ):
        """
        Runs the inner optimization dynamics for T iterations
        in the meanwhile.

        :param inner_objective_feed_dicts: Optional feed dictionary for the inner objective
        :param outer_objective_feed_dicts: Optional feed dictionary for the outer objective
                                            (note that this is not used in ForwardHG since hypergradients are not
                                            variables)
        :param initializer_feed_dict: Optional feed dictionary for the inner objective
        :param global_step: Optional global step for the optimization process
        :param param_dict: dictionary of parameters passed by `boml.boml_optimizer.BOMLOptimizer`
        :param session: Optional session (otherwise will take the default session)
        """
        raise NotImplementedError()

    def hgrads_hvars(self, meta_param=None, aggregation_fn=None, gradient_clip=None):
        """
        Method for getting outergradient and outer parameters as required by apply_gradient methods from tensorflow
        optimizers.

        :param meta_param: Optional list of outer parameters to consider. If not provided will get all variables in the
                            hyperparameter collection in the current scope.
        :param aggregation_fn: Optional operation to aggregate multiple hypergradients (for the same hyperparameter),
                                by default reduce_mean
        :param gradient_clip: Optional operation like clipping to be applied.
        :return:
        """
        if meta_param is None:
            meta_param = boml.extension.meta_parameters(tf.get_variable_scope().name)

        assert all(
            [h in self._hypergrad_dictionary for h in meta_param]
        ), "FINAL ERROR!"

        if aggregation_fn is None:
            aggregation_fn = lambda hgrad_list: tf.reduce_mean(hgrad_list, axis=0)

        def _aggregate_process_manage_collection(_hg_lst):
            if len(_hg_lst) == 1:  # avoid useless operations...
                aggr = _hg_lst[0]
            else:
                with tf.name_scope(_hg_lst[0].op.name):
                    aggr = aggregation_fn(_hg_lst) if len(_hg_lst) > 1 else _hg_lst[0]
            if gradient_clip is not None:
                with tf.name_scope("process_gradients"):
                    aggr = gradient_clip(aggr)
            tf.add_to_collection(boml.extension.GraphKeys.OUTERGRADIENTS, aggr)
            return aggr

        return [
            (_aggregate_process_manage_collection(self._hypergrad_dictionary[h]), h)
            for h in meta_param
        ]

    @property
    def name(self):
        return self._name

    def __str__(self):
        return self._name



from boml.extension import remove_from_collection
from boml.setup_model import network_utils
from boml.setup_model.network import *
from boml.setup_model.network_utils import as_tuple_or_list


class BOMLNetMiniMetaInitV2(BOMLNet):
    def __init__(
        self,
        _input,
        dim_output,
        name="BOMLNetMiniMetaInitV2",
        outer_param_dict=OrderedDict(),
        model_param_dict=OrderedDict(),
        task_parameter=OrderedDict(),
        activation=tf.nn.relu,
        var_collections=tf.GraphKeys.MODEL_VARIABLES,
        outer_method="Simple",
        conv_initializer=tf.contrib.layers.xavier_initializer_conv2d(tf.float32),
        output_weight_initializer=tf.contrib.layers.xavier_initializer(tf.float32),
        batch_norm=True,
        data_type=tf.float32,
        channels=3,
        dim_resnet=[64, 96, 128, 256],
        kernel=3,
        reuse=False,
        use_T=False,
        use_Warp=False,
    ):
        self.task_parameter = task_parameter
        self.dims = as_tuple_or_list(dim_output)
        self.kernel = kernel
        self.dim_resnet = dim_resnet
        self.channels = channels
        self.datatype = data_type
        self.activation = activation
        self.batch_norm = batch_norm
        self.stride = [1, 2, 2, 1]
        self.no_stride = [1, 1, 1, 1]
        self.conv_initializer = conv_initializer
        self.output_weight_initializer = output_weight_initializer
        self.use_T = use_T
        self.outer_method = outer_method
        self.use_Warp = use_Warp

        super().__init__(
            _input=_input,
            outer_param_dict=outer_param_dict,
            model_param_dict=model_param_dict,
            var_collections=var_collections,
            name=name,
            reuse=reuse,
        )
        # variables from batch normalization
        self.betas = self.filter_vars("beta")
        # moving mean and variance (these variables should be used at inference time... so must save them)
        self.moving_means = self.filter_vars("moving_mean")
        self.moving_variances = self.filter_vars("moving_variance")

        if not reuse:  # these calls might print a warning... it's not a problem..
            remove_from_collection(GraphKeys.MODEL_VARIABLES, *self.betas)
            remove_from_collection(GraphKeys.MODEL_VARIABLES, *self.moving_means)
            remove_from_collection(GraphKeys.MODEL_VARIABLES, *self.moving_variances)
        print(name, "MiniImagenet_MODEL CREATED")

    def create_outer_parameters(self, var_collections=GraphKeys.METAPARAMETERS):
        for i in range(len(self.dim_resnet)):
            self.outer_param_dict["res" + str(i + 1) + "id"] = tf.get_variable(
                name="res" + str(i + 1) + "id",
                shape=[
                    1,
                    1,
                    self.channels if i == 0 else self.dim_resnet[i - 1],
                    self.dim_resnet[i],
                ],
                dtype=self.datatype,
                initializer=self.conv_initializer,
            )
            for j in range(3):
                if i == 0:
                    self.outer_param_dict[
                        "res" + str(i + 1) + "conv_w" + str(j + 1)
                    ] = tf.get_variable(
                        name="res" + str(i + 1) + "conv_w" + str(j + 1),
                        shape=[
                            3,
                            3,
                            (self.channels if j == 0 else self.dim_resnet[i]),
                            self.dim_resnet[i],
                        ],
                        dtype=self.datatype,
                        initializer=self.conv_initializer,
                    )
                else:
                    self.outer_param_dict[
                        "res" + str(i + 1) + "conv_w" + str(j + 1)
                    ] = tf.get_variable(
                        name="res" + str(i + 1) + "conv_w" + str(j + 1),
                        shape=[
                            3,
                            3,
                            (self.dim_resnet[i - 1] if j == 0 else self.dim_resnet[i]),
                            self.dim_resnet[i],
                        ],
                        dtype=self.datatype,
                        initializer=self.conv_initializer,
                    )

        self.outer_param_dict["res_conv1"] = tf.get_variable(
            name="conv1",
            shape=[1, 1, self.dim_resnet[-1], 2048],
            initializer=self.conv_initializer,
            dtype=self.datatype,
        )
        self.outer_param_dict["res_bias1"] = tf.Variable(
            tf.zeros([2048]), name="res_bias1"
        )
        self.outer_param_dict["res_conv2"] = tf.get_variable(
            name="res_conv2",
            dtype=self.datatype,
            shape=[1, 1, 2048, 512],
            initializer=self.conv_initializer,
        )
        self.outer_param_dict["res_bias2"] = tf.Variable(
            tf.zeros([512]), name="res_bias2"
        )
        self.outer_param_dict["fully_connected_weights"] = tf.get_variable(
            name="fully_connected_weights",
            dtype=self.datatype,
            shape=[512, self.dims[-1]],
            initializer=self.output_weight_initializer,
        )
        self.outer_param_dict["fully_connected_bias"] = tf.Variable(
            name="fully_connected_bias", initial_value=tf.zeros([self.dims[-1]])
        )
        [
            tf.add_to_collection(var_collections, hyper)
            for hyper in self.outer_param_dict.values()
        ]
        return self.outer_param_dict

    def _forward(self):
        if len(self.task_parameter) == 0:
            self.task_parameter = self.create_initial_parameter(
                primary_outerparameter=self.outer_param_dict
            )

        def residual_block(x, i):
            skipc = tf.nn.conv2d(
                input=x,
                filter=self.task_parameter["res" + str(i + 1) + "id"],
                strides=self.no_stride,
                padding="SAME",
            )

            def conv_block(xx, i, j):
                out = tf.nn.conv2d(
                    xx,
                    self.task_parameter["res" + str(i + 1) + "conv_w" + str(j + 1)],
                    self.no_stride,
                    "SAME",
                )
                out = tf.contrib.layers.batch_norm(out, activation_fn=None)
                return network_utils.leaky_relu(out, 0.1)

            out = x
            for j in range(3):
                out = conv_block(out, i, j)

            add = tf.add(skipc, out)
            return tf.nn.max_pool(add, [1, 2, 2, 1], [1, 2, 2, 1], "SAME")

        for i in range(len(self.dim_resnet)):
            self + residual_block(self.out, i)
        self + tf.add(
            tf.nn.conv2d(
                self.out, self.task_parameter["res_conv1"], self.no_stride, "SAME"
            ),
            self.task_parameter["res_bias1"],
        )
        self + tf.nn.avg_pool(self.out, [1, 6, 6, 1], [1, 6, 6, 1], "VALID")
        self + tf.add(
            tf.nn.conv2d(
                self.out, self.task_parameter["res_conv2"], self.no_stride, "SAME"
            ),
            self.task_parameter["res_bias2"],
        )
        self + tf.reshape(self.out, (-1, 512))
        self + tf.add(
            tf.matmul(self.out, self.task_parameter["fully_connected_weights"]),
            self.task_parameter["fully_connected_bias"],
        )

    def re_forward(self, new_input=None, task_parameter=OrderedDict()):
        return BOMLNetMiniMetaInitV2(
            _input=new_input if (new_input is not None) else self.layers[0],
            dim_output=self.dims[-1],
            name=self.name,
            activation=self.activation,
            outer_method=self.outer_method,
            outer_param_dict=self.outer_param_dict,
            model_param_dict=self.model_param_dict,
            task_parameter=self.task_parameter
            if len(task_parameter.keys()) == 0
            else task_parameter,
            var_collections=self.var_collections,
            output_weight_initializer=self.output_weight_initializer,
            reuse=True,
            use_T=self.use_T,
            use_Warp=self.use_Warp,
        )


class BOMLNetOmniglotMetaInitV2(BOMLNet):
    def __init__(
        self,
        _input,
        dim_output,
        name="Omniglot_ResNet",
        outer_param_dict=OrderedDict(),
        model_param_dict=OrderedDict(),
        task_parameter=OrderedDict(),
        activation=tf.nn.relu,
        var_collections=tf.GraphKeys.MODEL_VARIABLES,
        conv_initializer=tf.contrib.layers.xavier_initializer_conv2d(tf.float32),
        outer_method="Simple",
        output_weight_initializer=tf.zeros_initializer(tf.float32),
        batch_norm=True,
        data_type=tf.float32,
        channels=1,
        dim_resnet=[64, 96],
        dim_hidden=64,
        kernel=3,
        max_pool=False,
        reuse=False,
        use_T=False,
        use_Warp=False,
    ):
        self.task_parameter = task_parameter
        self.kernel = kernel
        self.channels = channels
        self.dim_resnet = dim_resnet
        self.dims = as_tuple_or_list(dim_output)
        self.dim_hidden = dim_hidden
        self.datatype = data_type
        self.batch_norm = batch_norm
        self.max_pool = max_pool
        self.stride = [1, 2, 2, 1]
        self.no_stride = [1, 1, 1, 1]
        self.activation = activation
        self.conv_initializer = conv_initializer
        self.output_weight_initializer = output_weight_initializer
        self.use_T = use_T
        self.outer_method = (outer_method,)
        self.use_Warp = use_Warp
        super().__init__(
            _input=_input,
            outer_param_dict=outer_param_dict,
            model_param_dict=model_param_dict,
            var_collections=var_collections,
            name=name,
            reuse=reuse,
        )

        # variables from batch normalization
        self.betas = self.filter_vars("beta")
        # moving mean and variance (these variables should be used at inference time... so must save them)
        self.moving_means = self.filter_vars("moving_mean")
        self.moving_variances = self.filter_vars("moving_variance")

        if not reuse:  # these calls might print a warning... it's not a problem..
            remove_from_collection(GraphKeys.MODEL_VARIABLES, *self.betas)
            remove_from_collection(GraphKeys.MODEL_VARIABLES, *self.moving_means)
            remove_from_collection(GraphKeys.MODEL_VARIABLES, *self.moving_variances)
            print(name, "MODEL CREATED")

    def create_outer_parameters(self, var_collections=GraphKeys.METAPARAMETERS):
        for i in range(len(self.dim_resnet)):
            self.outer_param_dict["res" + str(i + 1) + "id"] = tf.get_variable(
                name="res" + str(i + 1) + "id",
                shape=[
                    1,
                    1,
                    self.channels if i == 0 else self.dim_resnet[i - 1],
                    self.dim_resnet[i],
                ],
                dtype=self.datatype,
                initializer=self.conv_initializer,
            )
            for j in range(3):
                if i == 0:
                    self.outer_param_dict[
                        "res" + str(i + 1) + "conv_w" + str(j + 1)
                    ] = tf.get_variable(
                        name="res" + str(i + 1) + "conv_w" + str(j + 1),
                        shape=[
                            3,
                            3,
                            (self.channels if j == 0 else self.dim_resnet[i]),
                            self.dim_resnet[i],
                        ],
                        dtype=self.datatype,
                        initializer=self.conv_initializer,
                    )
                else:
                    self.outer_param_dict[
                        "res" + str(i + 1) + "conv_w" + str(j + 1)
                    ] = tf.get_variable(
                        name="res" + str(i + 1) + "conv_w" + str(j + 1),
                        shape=[
                            3,
                            3,
                            (self.dim_resnet[i - 1] if j == 0 else self.dim_resnet[i]),
                            self.dim_resnet[i],
                        ],
                        dtype=self.datatype,
                        initializer=self.conv_initializer,
                    )
                self.outer_param_dict[
                    "res" + str(i + 1) + "conv_bias" + str(j + 1)
                ] = tf.get_variable(
                    name="res" + str(i + 1) + "conv_bias" + str(j + 1),
                    shape=[self.dim_resnet[i]],
                    initializer=tf.zeros_initializer(tf.float32),
                    dtype=self.datatype,
                )
        self.outer_param_dict["res_conv1"] = tf.get_variable(
            name="conv1",
            shape=[1, 1, self.dim_resnet[-1], 2048],
            initializer=self.conv_initializer,
            dtype=self.datatype,
        )
        self.outer_param_dict["res_bias1"] = tf.Variable(
            tf.zeros([2048]), name="res_bias1"
        )
        self.outer_param_dict["res_conv2"] = tf.get_variable(
            name="res_conv2",
            dtype=self.datatype,
            shape=[1, 1, 2048, 512],
            initializer=self.conv_initializer,
        )
        self.outer_param_dict["res_bias2"] = tf.Variable(
            tf.zeros([512]), name="res_bias2"
        )
        self.outer_param_dict["fully_connected_weights"] = tf.get_variable(
            name="fully_connected_weights",
            dtype=self.datatype,
            shape=[512, self.dims[-1]],
            initializer=self.output_weight_initializer,
        )
        self.outer_param_dict["fully_connected_bias"] = tf.Variable(
            name="fully_connected_bias", initial_value=tf.zeros([self.dims[-1]])
        )
        [
            tf.add_to_collection(var_collections, hyper)
            for hyper in self.outer_param_dict.values()
        ]
        return self.outer_param_dict

    def _forward(self):
        if len(self.task_parameter) == 0:
            self.task_parameter = self.create_initial_parameter(
                primary_outerparameter=self.outer_param_dict
            )

        for i in range(len(self.dim_resnet)):
            self + self.residual_block(self.out, i)

        self + tf.add(
            tf.nn.conv2d(
                self.out, self.task_parameter["res_conv1"], self.no_stride, "SAME"
            ),
            self.task_parameter["res_bias1"],
        )
        self + tf.nn.avg_pool(self.out, [1, 6, 6, 1], [1, 6, 6, 1], "VALID")
        self + tf.add(
            tf.nn.conv2d(
                self.out, self.task_parameter["res_conv2"], self.no_stride, "SAME"
            ),
            self.task_parameter["res_bias2"],
        )
        self + tf.reshape(self.out, (-1, 512))
        self + tf.add(
            tf.matmul(self.out, self.task_parameter["fully_connected_weights"]),
            self.task_parameter["fully_connected_bias"],
        )

    def residual_block(self, x, i):
        skipc = tf.nn.conv2d(
            x, self.task_parameter["res" + str(i + 1) + "id"], self.no_stride, "SAME"
        )
        out = x
        for j in range(3):
            out = self.conv_block(out, i, j)

        add = tf.add(skipc, out)
        return tf.nn.max_pool(add, [1, 2, 2, 1], [1, 2, 2, 1], "SAME")

    def conv_block(self, xx, i, j):
        out = tf.add(
            tf.nn.conv2d(
                xx,
                self.task_parameter["res" + str(i + 1) + "conv_w" + str(j + 1)],
                self.no_stride,
                "SAME",
            ),
            self.task_parameter["res" + str(i + 1) + "conv_bias" + str(j + 1)],
        )
        out = tf.contrib.layers.batch_norm(
            out,
            activation_fn=None,
            variables_collections=self.var_collections,
            scope="scope" + str(i + 1) + str(j + 1),
            reuse=self.reuse,
        )
        return network_utils.leaky_relu(out, 0.1)

    def re_forward(self, new_input=None, task_parameter=OrderedDict()):
        return BOMLNetOmniglotMetaInitV2(
            _input=new_input if new_input is not None else self.layers[0],
            dim_output=self.dims[-1],
            name=self.name,
            activation=self.activation,
            outer_param_dict=self.outer_param_dict,
            model_param_dict=self.model_param_dict,
            task_parameter=task_parameter
            if len(task_parameter) == 0
            else self.task_parameter,
            var_collections=self.var_collections,
            outer_method=self.outer_method,
            output_weight_initializer=self.output_weight_initializer,
            reuse=True,
            use_T=self.use_T,
            use_Warp=self.use_Warp,
        )



from collections import OrderedDict

import tensorflow as tf
from tensorflow.contrib import layers as tcl

import boml.extension
from boml.setup_model import network_utils
from boml.setup_model.network import BOMLNet


class BOMLNetMiniMetaReprV2(BOMLNet):
    def __init__(
        self,
        _input,
        name="BOMLNetMiniMetaReprV2",
        outer_param_dict=OrderedDict(),
        dim_output=-1,
        model_param_dict=OrderedDict(),
        task_parameter=OrderedDict(),
        use_T=False,
        use_Warp=False,
        reuse=False,
        outer_method="Reverse",
    ):
        self.var_coll = boml.extension.METAPARAMETERS_COLLECTIONS
        self.task_paramter = task_parameter
        self.outer_method = outer_method
        self.dim_output = dim_output
        self.use_T = use_T
        self.use_Warp = use_Warp

        super().__init__(
            _input=_input,
            outer_param_dict=outer_param_dict,
            model_param_dict=model_param_dict,
            name=name,
            reuse=reuse,
        )

        self.betas = self.filter_vars("beta")
        self.moving_means = self.filter_vars("moving_mean")
        self.moving_variances = self.filter_vars("moving_variance")
        if not reuse:
            boml.extension.remove_from_collection(
                boml.extension.GraphKeys.MODEL_VARIABLES,
                *self.moving_means,
                *self.moving_variances
            )

        boml.extension.remove_from_collection(
            boml.extension.GraphKeys.METAPARAMETERS,
            *self.moving_means,
            *self.moving_variances
        )
        print(name, "MODEL CREATED")

    def _forward(self):
        def residual_block(x, n_filters):
            skip_c = tcl.conv2d(x, n_filters, 1, activation_fn=None)

            def conv_block(xx):
                out = tcl.conv2d(
                    xx,
                    n_filters,
                    3,
                    activation_fn=None,
                    normalizer_fn=tcl.batch_norm,
                    variables_collections=self.var_coll,
                )
                return network_utils.leaky_relu(out, 0.1)

            out = x
            for _ in range(3):
                out = conv_block(out)

            add = tf.add(skip_c, out)

            return tf.nn.max_pool(add, [1, 2, 2, 1], [1, 2, 2, 1], "SAME")

        self + residual_block(self.out, 64)
        self + residual_block(self.out, 96)
        self + residual_block(self.out, 128)
        self + residual_block(self.out, 256)
        self + tcl.conv2d(self.out, 2048, 1, variables_collections=self.var_coll)
        self + tf.nn.avg_pool(self.out, [1, 6, 6, 1], [1, 6, 6, 1], "VALID")
        self + tcl.conv2d(self.out, 512, 1, variables_collections=self.var_coll)
        self + tf.reshape(self.out, (-1, 512))

    def re_forward(self, new_input=None):
        return BOMLNetMiniMetaReprV2(
            _input=new_input if new_input is not None else self.layers[0],
            model_param_dict=self.model_param_dict,
            name=self.name,
            dim_output=self.dim_output,
            outer_param_dict=self.outer_param_dict,
            reuse=True,
            outer_method=self.outer_method,
            use_T=self.use_T,
        )


class BOMLNetOmniglotMetaReprV2(BOMLNet):
    def __init__(
        self,
        _input,
        name="BOMLNetOmniglotMetaReprV2",
        outer_param_dict=OrderedDict(),
        dim_output=-1,
        model_param_dict=OrderedDict(),
        use_T=False,
        use_Warp=False,
        reuse=False,
        outer_method="Reverse",
    ):
        self.var_coll = boml.extension.METAPARAMETERS_COLLECTIONS
        self.outer_method = outer_method
        self.dim_output = dim_output
        self.use_T = use_T
        self.use_Warp = use_Warp
        super().__init__(
            _input=_input,
            outer_param_dict=outer_param_dict,
            model_param_dict=model_param_dict,
            name=name,
            reuse=reuse,
        )

        self.betas = self.filter_vars("beta")
        self.moving_means = self.filter_vars("moving_mean")
        self.moving_variances = self.filter_vars("moving_variance")

        if not reuse:
            boml.extension.remove_from_collection(
                boml.extension.GraphKeys.MODEL_VARIABLES,
                *self.moving_means,
                *self.moving_variances
            )

        boml.extension.remove_from_collection(
            boml.extension.GraphKeys.METAPARAMETERS,
            *self.moving_means,
            *self.moving_variances
        )
        print(name, "MODEL CREATED")

    def _forward(self):
        def residual_block(x, n_filters):
            skip_c = tcl.conv2d(x, n_filters, 1, activation_fn=None)

            def conv_block(xx):
                out = tcl.conv2d(
                    xx,
                    n_filters,
                    3,
                    activation_fn=None,
                    normalizer_fn=tcl.batch_norm,
                    variables_collections=self.var_coll,
                )
                return network_utils.leaky_relu(out, 0.1)

            out = x
            for _ in range(3):
                out = conv_block(out)

            add = tf.add(skip_c, out)

            return tf.nn.max_pool(add, [1, 2, 2, 1], [1, 2, 2, 1], "SAME")

        self + residual_block(self.out, 64)
        self + residual_block(self.out, 96)
        self + tcl.conv2d(self.out, 2048, 1, variables_collections=self.var_coll)
        self + tf.nn.avg_pool(self.out, [1, 6, 6, 1], [1, 6, 6, 1], "VALID")
        self + tcl.conv2d(self.out, 512, 1, variables_collections=self.var_coll)
        self + tf.reshape(self.out, (-1, 512))

    def re_forward(self, new_input=None, task_parameter=OrderedDict()):
        return BOMLNetOmniglotMetaReprV2(
            new_input if new_input is not None else self.layers[0],
            model_param_dict=self.model_param_dict,
            name=self.name,
            dim_output=self.dim_output,
            outer_param_dict=self.outer_param_dict,
            reuse=True,
            use_T=self.use_T,
            outer_method=self.outer_method,
        )
